/*
//
//                  INTEL CORPORATION PROPRIETARY INFORMATION
//     This software is supplied under the terms of a license agreement or
//     nondisclosure agreement with Intel Corporation and may not be copied
//     or disclosed except in accordance with the terms of that agreement.
//          Copyright(c) 2007-2008 Intel Corporation. All Rights Reserved.
//
//     Intel(R) Integrated Performance Primitives Data Compression Domain-based GZIP
//
//  By downloading and installing this sample, you hereby agree that the
//  accompanying Materials are being provided to you under the terms and
//  conditions of the End User License Agreement for the Intel(R) Integrated
//  Performance Primitives product previously accepted by you. Please refer
//  to the file ippEULA.rtf located in the root directory of your Intel(R) IPP
//  product installation for more information.
//
*/
#include <stdio.h>
#include <stdlib.h>
#ifndef GZIP_PTHREADS
    #if defined(_OPENMP) && defined (GZIP_PARALLEL)
#include <omp.h>
#endif
#endif
#ifdef _WIN32
#include <windows.h>
#endif

#ifdef GZIP_PTHREADS
    #include <pthread.h>
#endif

#include "version.h"
#include "ipp_gzip.h"

/* Global variables */
short opt_ascii = 0;
short opt_to_stdout = 0;
short opt_decompress = 0;
short opt_encrypt = 0;
short opt_force = 0;
short opt_help = 0;
short opt_pkzip = 0;
short opt_list = 0;
short opt_license = 0;
short opt_noname = 0;
short opt_name = 0;
short opt_quiet = 0;
short opt_reqursive = 0;
short opt_new_suffix = 0;
short opt_test = 0;
short opt_notime = 0;
short opt_verbose = 0;
short opt_version = 0;
short opt_compress_fast = 0;
short opt_compress_best = 0;
short opt_lzw_compat = 0;
short opt_bits = 0;
short opt_rsyncable = 0;
short opt_force_dynamic = 0;
char *prog_name = NULL;
char *suffix = NULL;

short opt_dbg_multithreads = 0;
char *opt_num_threads_forced = NULL;
short opt_dbg_slice_size = 0;
char *opt_slice_size_array = NULL;
long opt_slice_size = 0;
short from_stdin = 0;
short opt_perf_stat = 0;

/* Input/output options storage */
short opt_io_read_method = METHOD_LOW;
char *ptr_read_method = NULL;
short opt_io_write_method = METHOD_MMAP;
char *ptr_write_method = NULL;
short opt_io_read_buffer = 64;
size_t INBUF_SIZE = 64 * 1024;
char *ptr_read_buffer = NULL;
short opt_io_write_buffer = 64;
size_t OUTBUF_SIZE = 64 * 1024;
char *ptr_write_buffer = NULL;
short opt_debug = 0;

#ifdef _WIN32
    SYSTEM_INFO sysInfo;
#else
    size_t sys_page_size;
#endif

/* Local variables */
int num_threads = 1;
#ifdef GZIP_PTHREADS
static pthread_t *pthreads;
static pthread_attr_t attr;
pthread_cond_t condition_cond = PTHREAD_COND_INITIALIZER;
pthread_mutex_t condition_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_t f_pthread;                                     /* Finishing pthread identifier */
void *process_file_st(void*);
#else
static void process_file_st(char*);
#endif
static void process_file_mt(char*);

int main(int argc, char *argv[])
{
    int i;
    int file_arg_start;
    int file_count;

    prog_name = get_program_name(argv[0]);
    /* Simple option parser */
    file_arg_start = get_options(argc, argv);
    file_count = argc - file_arg_start;
    if (opt_help) {
        help();
        exit(0);
    }
    if (opt_version) {
        print_version();
    }
    if (opt_ascii && !opt_quiet) {
        fprintf(stderr, "%s: 'ascii' is ignored in this version\n", prog_name);
    }
    if (opt_encrypt && !opt_quiet) {
        fprintf(stderr, "%s: 'encrypt' is ignored in this version\n", prog_name);
    }
    if (opt_pkzip && !opt_quiet) {
        fprintf(stderr, "%s: 'pkzip' is ignored in this version\n", prog_name);
    }
    /* Decode input/output options */
#ifdef _WIN32
    GetSystemInfo(&sysInfo);            /* Get OS file map granularity */
#else
    sys_page_size = (size_t)sysconf(_SC_PAGESIZE);
#endif
    if (ptr_read_buffer != NULL) {      /* Read buffer size is specified */
        opt_io_read_buffer = (short)atoi(ptr_read_buffer);
        if (opt_io_read_buffer == 0 || opt_io_read_buffer > 1024)    /* If something wrong is specified */
            opt_io_read_buffer = 64;
        INBUF_SIZE = opt_io_read_buffer * 1024;
    }
    if (ptr_write_buffer != NULL) {
        opt_io_write_buffer = (short)atoi(ptr_write_buffer);
        if (opt_io_write_buffer == 0 || opt_io_write_buffer > 1024) 
            opt_io_write_buffer = 64;
        OUTBUF_SIZE = opt_io_write_buffer * 1024;
    }
    if (ptr_read_method != NULL) {
        opt_io_read_method = (short)atoi(ptr_read_method);
        if (opt_io_read_method < METHOD_LOW || opt_io_read_method > METHOD_MMAP)
            opt_io_read_method = METHOD_LOW;         /* Low/level by default */
        if (opt_io_read_method == METHOD_MMAP) {
#ifdef _WIN32
            /* Align buffer size to system granularity boundary */
            INBUF_SIZE = (INBUF_SIZE/sysInfo.dwAllocationGranularity+1)*sysInfo.dwAllocationGranularity;
#else
            INBUF_SIZE = (INBUF_SIZE/sys_page_size + 1) * sys_page_size;
#endif
        }
    }
    if (ptr_write_method != NULL) {
        opt_io_write_method = (short)atoi(ptr_write_method);
        if (opt_io_write_method < METHOD_LOW || opt_io_write_method > METHOD_MMAP) 
            opt_io_write_method = METHOD_MMAP;        /* mmap by default */
        if (opt_io_write_method == METHOD_MMAP) {
#ifdef _WIN32
            /* Align buffer size to system granularity boundary */
            OUTBUF_SIZE = (OUTBUF_SIZE/sysInfo.dwAllocationGranularity+1)*sysInfo.dwAllocationGranularity;
#else
            OUTBUF_SIZE = (OUTBUF_SIZE/sys_page_size + 1) * sys_page_size;
#endif
        }
    }
    if (file_count == 0) {
        if (!opt_force && !opt_test && !opt_list) {
            warn_message("compressed data not written to a terminal. Use -f to force compression.\n");
            fprintf(stderr, "For help, type: %s -h", prog_name);
            exit(0);
        }
        from_stdin = 1;
        opt_to_stdout = 1;
        opt_noname = 1;
        opt_notime = 1;
    }
    /* Determine number of available threads. Depending on it we'll choose between single thread and multi thread processing */
    if (opt_dbg_multithreads) {
        int nt = atoi(opt_num_threads_forced);
        if (nt > MAX_THREADS) {
            warn_message("Max. thread number exceeded");
            nt = MAX_THREADS;
        }
        if (nt > 0) {
#ifndef GZIP_PTHREADS
    #if defined(_OPENMP) && defined(GZIP_PARALLEL)
    #ifndef _WIN32
    #define _MSC_VER 0
    #endif
    #if (_MSC_VER >= 1400  && !defined(_AMD64_)) || defined __INTEL_COMPILER
                omp_set_num_threads(nt);
    #endif
    #else
                nt = 1;
    #endif
#endif
            num_threads = nt;
        }
    } else {
        num_threads = get_cpu_count();
    }
    if (opt_dbg_slice_size)
        opt_slice_size = atol(opt_slice_size_array);
    if (file_count > 1) {           /* Use multi-file parallel processing */
        if (num_threads > 1) {
            int act_threads;
#ifdef GZIP_PTHREADS
            int next_thread = 0;
#endif

#ifndef GZIP_PTHREADS
#if defined(__INTEL_COMPILER) && defined(_OPENMP) && defined(GZIP_PARALLEL)
            kmp_set_library_throughput();           /* Makes parallel thread to stop immediately after function execution */
            kmp_set_blocktime(0);                   /* Parallel threads will wait for 5 ms before stopping */
#endif
#endif
            if (file_count > num_threads) {
                act_threads = num_threads;
            } else {
                act_threads = file_count;
            }
#ifdef GZIP_PTHREADS
            pthreads = (pthread_t*)zalloc(act_threads * sizeof(pthread_t));
            pthread_attr_init(&attr);
            pthread_cond_init(&condition_cond, NULL);
            for (i=0; i<file_count; i++) {
                if (next_thread <= act_threads-1) {  /* here's free threads to use */
                    pthread_create(&pthreads[next_thread++], &attr, process_file_st, (void*)argv[file_arg_start+i]);
                } else {                            /* no free threads are available. wait for some of threads to finish */
                    int j;
                    pthread_mutex_lock(&condition_mutex);
                    pthread_cond_wait(&condition_cond, &condition_mutex);   /* Wait for any pthread to finish */
                    pthread_join(f_pthread, NULL);                          /* f_pthread must contain identifier of finishing pthread */
                    for (j=0; j<act_threads; j++) {
                        if (pthread_equal(f_pthread, pthreads[j])) {         /* This thread finished. start new pthread */
                            pthread_create(&pthreads[j], &attr, process_file_st, (void*)argv[file_arg_start+i]);
                            break;
                        }
                    }
                    pthread_mutex_unlock(&condition_mutex);                 /* Free other pthreads to finish */
                }
            }
            /* Wait for all threads finish */
            for (i=0; i<act_threads; i++) {
                pthread_join(pthreads[i], NULL);
            }
            pthread_attr_destroy(&attr);
            zfree(pthreads);
#else
#if defined(_OPENMP) && defined(GZIP_PARALLEL)
#pragma omp parallel for num_threads(act_threads)
#endif
            for (i=0; i<file_count; i++) {
                process_file_st(argv[file_arg_start + i]);
            }
#endif
        } else {
            for (i=0; i<file_count; i++) {
                process_file_st(argv[file_arg_start + i]);
            }
        }
    } else {                        /* file_count==1. Use multi-chunk processing */
#if defined(GZIP_PTHREADS) || (defined(_OPENMP) && defined(GZIP_PARALLEL)) 
            process_file_mt(argv[file_arg_start]);
#else
            process_file_st(argv[file_arg_start]);
#endif
        }
    return 0;
}

/* 
    print_version()
        simply prints copyright message and current program version number
*/
void print_version()
{
    fprintf(stderr, "%s\n%s\n\n", COPYRIGHT_MESSAGE, CURRENT_VERSION);
    fprintf(stderr, "Target architecture: %s\n", TARGET_ARCHITECTURE);
    fprintf(stderr, "Optimization: %s\n", OPTIMIZATION);
    fprintf(stderr, "Library: %s\n\n", LIBRARY_LINK);
}
/*
    process_file(char *filename)
        function to process compress/decompress/list/test on a file
*/
#ifdef GZIP_PTHREADS
void *process_file_st(void *param)
{
    char *filename;
    filename = (char*)param;
#else
void process_file_st(char *filename)
{
#endif
    if (opt_decompress | opt_list | opt_test) {
        inflate_st(filename);
    } else
        deflate_st(filename);
    if (num_threads == 1)
#ifdef GZIP_PTHREADS
        return NULL;
    pthread_mutex_lock(&condition_mutex);       /* Block other threads from going further */
    f_pthread = pthread_self();                 /* Get self thread identifier */
    pthread_cond_signal(&condition_cond);       /* Signal to master thread to proceed */
    pthread_mutex_unlock(&condition_mutex);     /* Allow master thread to proceed */
    return NULL;
#else
        return;                         /* If single-thread, do nothing */
#endif
}

void process_file_mt(char *filename)
{
    if (opt_decompress | opt_list | opt_test) {
        inflate_mt(filename);
    } else
        deflate_mt(filename);
}

