/*
//
//                  INTEL CORPORATION PROPRIETARY INFORMATION
//     This software is supplied under the terms of a license agreement or
//     nondisclosure agreement with Intel Corporation and may not be copied
//     or disclosed except in accordance with the terms of that agreement.
//          Copyright(c) 2007-2008 Intel Corporation. All Rights Reserved.
//
//     Intel(R) Integrated Performance Primitives Data Compression Domain-based GZIP
//
//  By downloading and installing this sample, you hereby agree that the
//  accompanying Materials are being provided to you under the terms and
//  conditions of the End User License Agreement for the Intel(R) Integrated
//  Performance Primitives product previously accepted by you. Please refer
//  to the file ippEULA.rtf located in the root directory of your Intel(R) IPP
//  product installation for more information.
//
*/
#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include <fcntl.h>
#ifndef GZIP_PTHREADS
#if defined(_OPENMP) && defined(GZIP_PARALLEL)
    #include <omp.h>
#endif
#endif
#ifndef _WIN32
    #include <dirent.h>
    #include <unistd.h>
#endif

#include <ippcore.h>
#include <ippdc.h>

#include "ipp_gzip.h"

/*! \brief Struct to pack all local variables to simplify parameters passing */
struct deflate_data_pack {
    UBYTE *inbuf;
    UBYTE *outbuf;
    UBYTE *inptr;
    UBYTE *outptr;
    GZIP_FILE *fd_in;
    GZIP_FILE *fd_out;
    char *filename;
    long file_size;
    IppLZ77State_8u *ippState;
    IppLZ77Pair *p_pair;
    IppLZ77Pair *c_pair;
    size_t inlen;
    size_t outlen;
    int c_index;
/*! \brief Initial free elements number in pair array */
    int c_len;
/*! \brief Free elements number in pair array */
    int len;
};

static void stop_deflating(struct deflate_data_pack*);
static IppStatus encode_Huffman(IppLZ77HuffMode, IppLZ77Flush, struct deflate_data_pack*, int, UBYTE*, size_t);

/**
 * Signle-threaded version of "deflate" function
 * @param filename - name of file to compress
 */
#ifdef GZIP_PTHREADS
void *deflate_st(void* param)
#else
void *deflate_st(char *filename)
#endif
{
#if _MSC_VER >= 1400  && !defined(_AMD64_)
    struct _stat file_info;
#else
    struct stat file_info;
#endif
    char out_filename[MAX_NAME];
    char base_name[MAX_NAME];
    int incnt;      /* Last read operation will return this */
    IppStatus status;
    IppLZ77ComprLevel compression_level;
    IppLZ77Flush flush_mode;
    struct deflate_data_pack *pdata;
    short no_last_encode;
    Ipp64u stat_start_clocks, stat_finish_clocks;
    IppLZ77HuffMode huff_mode;
    UBYTE stored_block_save[STORED_BUF_SIZE];
    size_t stored_block_idx_low, stored_block_idx_high;
    size_t stored_block_watermark, stored_block_len; 
    int can_use_stored;
#ifdef GZIP_PTHREADS
    char *filename;

    filename = (char*)param;
#endif

    if (!from_stdin) {
        /* Skip bogus files */
        if (strcmp(".", filename) == 0 || strcmp("..", filename) == 0)
            return NULL;
        if (known_suffix(filename))
            return NULL;
    }
    /* Allocate data pack */
    pdata = (struct deflate_data_pack *) zalloc(sizeof(struct deflate_data_pack));
    if (!from_stdin) {
#if _MSC_VER >= 1400  && !defined(_AMD64_)
        if (_stat(filename, &file_info) == -1) {
#else
        if (stat(filename, &file_info) == -1) {
#endif
            error_file(filename);
            zfree(pdata);
    #ifdef GZIP_PTHREADS
            return NULL;
    #else
            return 0;
    #endif
        }
        /* Check if directory file is specified */
#if _MSC_VER >= 1400  && !defined(_AMD64_)
        if ((file_info.st_mode & _S_IFDIR) != 0) {
#else
        if ((file_info.st_mode & S_IFDIR) != 0) {
#endif
            if (opt_reqursive == 0) {
                warn_message("directory ignored");
                return 0;
            }
            return compress_dir_st(filename);       /* Process directory in single-threaded environment. 
                                                    If we are multi-CPU, the directory can be processed in parallel */
        }
        pdata->file_size = file_info.st_size;
    }   /* !from_stdin */
    /* Start CPU clocking if required */
    if (opt_perf_stat)
        stat_start_clocks = ippGetCpuClocks();
    /* Open input file */
    if (from_stdin)
        pdata->fd_in = zopen_stdin();
    else {
        pdata->fd_in = ZOPEN(filename, GZIP_FILE_READ, 0);
    }
    if (pdata->fd_in == NULL) {                  /* Cannot open input file for sharing for some reason */
        error_file(filename);
        zfree(pdata);
        return 0;
    }
    if (!opt_to_stdout) {
        create_output_name(filename, out_filename);
        if ((pdata->fd_out = out_open(out_filename, zgetsize(pdata->fd_in))) == NULL) {
            error_file(out_filename);
            zclose(pdata->fd_in);
            zfree(pdata);
            return 0;
        }
        pdata->fd_out->uncompressed_size = zgetsize(pdata->fd_in);
    } else
        pdata->fd_out = zopen_stdout();
    pdata->filename = out_filename;
    /* Allocate input/output buffers */
    pdata->inbuf = (UBYTE*) zalloc(INBUF_SIZE);

    pdata->outptr = pdata->outbuf = (UBYTE*) zalloc(OUTBUF_SIZE);
    if (!opt_noname) 
        get_base_name(filename, base_name);
    pdata->outptr = create_gzip_hdr(pdata->outptr, base_name, &file_info, 0);   /* Last 0 means that we need no comment here */
    zwrite(pdata->fd_out, pdata->outbuf, pdata->outptr - pdata->outbuf);
    pdata->outptr = pdata->outbuf;
    if (file_info.st_size == 0) {           /* Special case specific for GZIP */
        PUT_BYTE(pdata->outptr, 0x03);
        PUT_BYTE(pdata->outptr, 0x00);
        PUT_LONG(pdata->outptr, 0L);
        PUT_LONG(pdata->outptr, 0L);
        zwrite(pdata->fd_out, pdata->outbuf, pdata->outptr - pdata->outbuf);
        if (opt_verbose && filename != NULL && !opt_to_stdout) {
            fprintf(stderr, "%s:\t %6.1f%% -- replaced with %s\n", filename, 100*(1.-(float)pdata->fd_out->bytes_out/pdata->fd_in->bytes_in), out_filename);
        }
        zclose(pdata->fd_in);
        zclose(pdata->fd_out);
        zfree(pdata->inbuf);
        zfree(pdata->outbuf);
        if (!opt_to_stdout && filename != NULL) {
            /* Print deflating statistics if necessary */
            UNLINK(filename);
        }
        zfree(pdata);
        return NULL;
    }
    compression_level = IppLZ77AverageCompr;
    if (opt_compress_best) {
        compression_level = IppLZ77BestCompr;
    }
    if (opt_compress_fast) {
        compression_level = IppLZ77FastCompr;
    }
    /* Here we ready to start compression */
    /* Allocate IppState structure */
    status = ippsEncodeLZ77InitAlloc_8u(compression_level, IppLZ77CRC32, &pdata->ippState);
    status = ippsEncodeLZ77SetStatus_8u(IppLZ77StatusLZ77Process, pdata->ippState);
    status = ippsEncodeLZ77GetPairs_8u(&pdata->c_pair, &pdata->c_index, &pdata->c_len, pdata->ippState);
    pdata->len = pdata->c_len;
    pdata->p_pair = pdata->c_pair + pdata->c_index;
    pdata->outlen = (int)(OUTBUF_SIZE - (pdata->outptr-pdata->outbuf));
    no_last_encode = 0;
    stored_block_watermark = 0;
    can_use_stored = 1;
    while (!zeof(pdata->fd_in)) {
        incnt = (int)zread(pdata->fd_in, pdata->inbuf, INBUF_SIZE);
        if (incnt == -1) {                      /* Some error during read occurred */
            error_file(filename);
            stop_deflating(pdata);
            zfree(pdata);
            return 0;
        }
        pdata->file_size -= incnt;
        if (incnt < (int)INBUF_SIZE || pdata->file_size == 0) {               /* We're close to or at the end of file */
            flush_mode = IppLZ77FinishFlush;
        } else {
            flush_mode = IppLZ77NoFlush;
        }
        pdata->inptr = pdata->inbuf;
        pdata->inlen = incnt;
        /* outptr is modified by gzip header output function in the very first loop cycle, so we don't need to set it here */
        while (1) {
            stored_block_idx_low = pdata->inptr - pdata->inbuf;       /* Save low index value for possible stored block ops */
            status = ippsEncodeLZ77_8u(&pdata->inptr, (int*)&pdata->inlen, &pdata->p_pair, &pdata->len, flush_mode, pdata->ippState);
            stored_block_idx_high = pdata->inptr - pdata->inbuf;
            stored_block_len = stored_block_idx_high - stored_block_idx_low;
            if (can_use_stored && (stored_block_watermark+stored_block_len) < STORED_BUF_SIZE) {
                memcpy(&stored_block_save[stored_block_watermark], &(pdata->inbuf[stored_block_idx_low]), stored_block_len);
                stored_block_watermark += stored_block_len;
                can_use_stored = 1;
            } else {    /* Processed block is too long to be stored */
                stored_block_watermark = 0;
                can_use_stored = 0;
            }
            if (status == ippStsNoErr) {
                if (!zeof(pdata->fd_in)) {
                    break;
                }
                /* Start Huffman flushing process */
                pdata->p_pair = pdata->c_pair;
                pdata->len = pdata->c_len - pdata->len;
                status = ippsEncodeLZ77SetStatus_8u(IppLZ77StatusHuffProcess, pdata->ippState);
                status = ippsEncodeLZ77SelectHuffMode_8u(pdata->p_pair, pdata->len, &huff_mode, pdata->ippState);
                pdata->outlen = OUTBUF_SIZE - (pdata->outptr-pdata->outbuf);
                status = encode_Huffman(huff_mode, flush_mode, pdata, can_use_stored, stored_block_save, stored_block_watermark);
                if (status != ippStsNoErr) {
                    no_last_encode = 1;
                    break;
                }
                pdata->p_pair = pdata->c_pair + pdata->c_index;
                pdata->len = pdata->c_len;
            } else if (status == ippStsDstSizeLessExpected) {
                /* Start Huffman flushing process */
                pdata->p_pair = pdata->c_pair;
                pdata->len = pdata->c_len - pdata->len;
                status = ippsEncodeLZ77SetStatus_8u(IppLZ77StatusHuffProcess, pdata->ippState);
                pdata->outlen = OUTBUF_SIZE - (pdata->outptr-pdata->outbuf);
                status = ippsEncodeLZ77SelectHuffMode_8u(pdata->p_pair, pdata->len, &huff_mode, pdata->ippState);
                status = encode_Huffman(huff_mode, IppLZ77NoFlush, pdata, can_use_stored, stored_block_save, stored_block_watermark);
                if (status != ippStsNoErr) {
                    stop_deflating(pdata);
                    zfree(pdata);
                    return 0;
                }
                pdata->p_pair = pdata->c_pair + pdata->c_index;
                pdata->len = pdata->c_len;
                can_use_stored = 1;
                stored_block_watermark = 0;
            } else {
                stop_deflating(pdata);
                print_ipp_status(status);
                zfree(pdata);
                return 0;
            }
        } /* while */
    }
    if (!no_last_encode) {
        /* Start Huffman flushing process */
        pdata->p_pair = pdata->c_pair;
        pdata->len = pdata->c_len - pdata->len;
        status = ippsEncodeLZ77SetStatus_8u(IppLZ77StatusFinal, pdata->ippState);
        status = ippsEncodeLZ77SelectHuffMode_8u(pdata->p_pair, pdata->len, &huff_mode, pdata->ippState);
        status = encode_Huffman(huff_mode, IppLZ77FinishFlush, pdata, can_use_stored, stored_block_save, stored_block_watermark);
    }
    if (opt_perf_stat && filename != NULL && !opt_to_stdout) {
        stat_finish_clocks = ippGetCpuClocks();
        fprintf(stderr, "%s:\t deflating -- ", filename);
        print_io_methods();
        fprintf(stderr, " ST -- %6.1f clocks per input symbol\n", (float)(stat_finish_clocks - stat_start_clocks) / pdata->fd_in->bytes_in);
    }
    if (opt_verbose && filename != NULL && !opt_to_stdout) {
        fprintf(stderr, "%s:\t %6.1f%% -- replaced with %s\n", filename, 100*(1.-(float)pdata->fd_out->bytes_out/pdata->fd_in->bytes_in), out_filename);
    }
    stop_deflating(pdata);
    if (!opt_to_stdout && filename != NULL) {
        /* Print deflating statistics if necessary */
        UNLINK(filename);
    }
    zfree(pdata);
    return NULL;
}
/**
 *  Finishes deflate process
 *      closes i/o files
 *      frees IPP State structure
 * @param fd1 - file 1 to close
 * @param fd2 - file 2 to close
 */
static void stop_deflating(struct deflate_data_pack *pdata)
{
    zclose(pdata->fd_in);
    zclose(pdata->fd_out);
    zfree(pdata->inbuf);
    zfree(pdata->outbuf);
    ippsLZ77Free_8u(pdata->ippState);
}
/**
 * Encode (and write) pairs buffer using selected Huffman
 * algorithm
 * @param huff_mode - Huffman encoding mode to use
 * @param len - starting length of pair array
 * @param outbuf - initial output buffer pointer
 * @param outlen - current output buffer free size
 * @param outbufsiz - initial output buffer length
 * @param flush_mode - Huffman flushing mode
 * @param fout - output file
 * @param filename - output file name (to signal in case of
 *                 write error)
 * 
 * @return IppStatus
 */
#define MAX_STORED_BLOCK (64*1024-1)
static IppStatus encode_Huffman(IppLZ77HuffMode huff_mode, IppLZ77Flush flush_mode, struct deflate_data_pack *pdata, int can_store, UBYTE *stored_block, size_t stored_block_len)
{
    IppStatus status;
    int write_status;

    size_t before_len;
    size_t len_add;

    len_add = 0;

    while (1) {
        switch (huff_mode) {
        case IppLZ77UseFixed:
            before_len = pdata->outlen;
            status = ippsEncodeLZ77FixedHuff_8u(&pdata->p_pair, &pdata->len, &pdata->outptr, (int*)&pdata->outlen, flush_mode, pdata->ippState);
            if (status != ippStsDstSizeLessExpected) {
                len_add = 0;
            } else {
                len_add += before_len - pdata->outlen;
            }
            break;
        case IppLZ77UseStored:
            if (can_store) {
                before_len = stored_block_len;
                status = ippsEncodeLZ77StoredBlock_8u(&stored_block, (int*)&stored_block_len, &pdata->outptr, (int*)&pdata->outlen, flush_mode, pdata->ippState);
                if (status != ippStsDstSizeLessExpected) {
                    len_add = 0;
                } else {
                    len_add += before_len - pdata->outlen;
                }
            break;
            }
            /* If we cannot stored block - e.g. it's size is too big - we break into dynamic Huffman mode */
        case IppLZ77UseDynamic:
            before_len = pdata->outlen;
            status = ippsEncodeLZ77DynamicHuff_8u(&pdata->p_pair, &pdata->len, &pdata->outptr, (int*)&pdata->outlen, flush_mode, pdata->ippState);
            len_add += before_len - pdata->outlen;
            if (status != ippStsDstSizeLessExpected) {
                len_add = 0;
            }
            break;
        default:
            warn_message("unknown Huffman encoding mode");
            return ippStsErr;
        }
        if (status == ippStsNoErr) {
                if (pdata->outlen == 0) {
                    write_status = (int)zwrite(pdata->fd_out, pdata->outbuf, OUTBUF_SIZE - pdata->outlen);
                    if (write_status == -1) {
                        error_file(pdata->filename);
                        return ippStsErr;
                    }
                    pdata->outptr = pdata->outbuf;
                    pdata->outlen = OUTBUF_SIZE;
                }
                return status;
        } else if (status == ippStsStreamEnd) {
            if (pdata->outlen < 8) {                  /* Remaining length less than need for CRC+FILE_LEN*/
                write_status = (int)zwrite(pdata->fd_out, pdata->outbuf, OUTBUF_SIZE - pdata->outlen);
                if (write_status == -1) {
                    error_file(pdata->filename);
                    return ippStsErr;
                }
                pdata->outptr = pdata->outbuf;
                pdata->outlen = OUTBUF_SIZE;
            }
            status = ippsEncodeLZ77SetStatus_8u(IppLZ77StatusFinal, pdata->ippState);
            CHECK_STATUS(status);
            status = ippsEncodeLZ77Flush_8u(&pdata->outptr, (int*)&pdata->outlen, pdata->ippState);
            write_status = (int)zwrite(pdata->fd_out, pdata->outbuf, OUTBUF_SIZE - pdata->outlen);
            if (write_status == -1) {
                error_file(pdata->filename);
                return ippStsErr;
            }
            return ippStsStreamEnd;
        } else if (status != ippStsDstSizeLessExpected) {
            return status;
        }
        write_status = (int)zwrite(pdata->fd_out, pdata->outbuf, OUTBUF_SIZE - pdata->outlen);
        if (write_status == -1) {
            error_file(pdata->filename);
            return ippStsErr;
        }
        pdata->outptr = pdata->outbuf;
        pdata->outlen = OUTBUF_SIZE;
    }
}
/**
 * Compress specified directory in single-thread. If current
 * computer has multiple CPUs, processing can be done in
 * parallel.
 * Windows NT has no POSIX-like directory processing functions,
 * so it will be done using Windows API.
 * In Linux we'll use POSIX functions
 * @param dirname
 * 
 * @return int
 */
static char BOGUS_FILE_1[] = ".";
static char BOGUS_FILE_2[] = "..";
#ifdef GZIP_PTHREADS
void *compress_dir_st(void* param)
#else
void *compress_dir_st(char *dirname)
#endif
{
#ifdef _WIN32
    WIN32_FIND_DATA FindFileData;
    HANDLE FindHandle;
    BOOL NextFile;
    char *valid_search_name;
#if defined(_OPENMP) && defined(GZIP_PARALLEL)
    char **file_table;
    int table_ptr;
#endif

    valid_search_name = (char*)zalloc(strlen(dirname)+3);
    strcpy(valid_search_name, dirname);
    strcat(valid_search_name, "\\*");
    FindHandle = FindFirstFile(valid_search_name, &FindFileData);
    if (FindHandle == INVALID_HANDLE_VALUE) {           /* No files in directory */
        zfree(valid_search_name);
        return NULL;
    }
#if defined(_OPENMP) && defined(GZIP_PARALLEL)
    if (num_threads == 1 || omp_in_parallel()) {                             /* True single-thread environment */
#else
        {
#endif
        NextFile = TRUE;
        while (NextFile) {
            char *full_filename;
            if (strcmp(BOGUS_FILE_1, FindFileData.cFileName) != 0 && strcmp(BOGUS_FILE_2, FindFileData.cFileName) != 0) {
                full_filename = (char*)zalloc(strlen(dirname)+strlen(FindFileData.cFileName)+2);
                strcpy(full_filename, dirname);
                strcat(full_filename, "\\");
                strcat(full_filename, FindFileData.cFileName);
                deflate_st(full_filename);
                zfree(full_filename);
            }
            NextFile = FindNextFile(FindHandle, &FindFileData);
        }
        if (GetLastError() == ERROR_NO_MORE_FILES) {
            FindClose(FindHandle);
            zfree(valid_search_name);
            return NULL;
        }
        perror("directory read failure");
        FindClose(FindHandle);
        zfree(valid_search_name);
        return 0;
#if !defined(_OPENMP) || !defined(GZIP_PARALLEL)
        }
#endif
#if defined(_OPENMP) && defined(GZIP_PARALLEL)
    } else {                                    /* More complex case - compressing directory in parallel */
        /* We'll be doing the following:
            read "num_threads" files names
            start processing of "num_threads" files
            read next portion */
        file_table = (char**)zalloc(num_threads * sizeof(char*));
        table_ptr = 0;
        while (1) {
            if (strcmp(BOGUS_FILE_1, FindFileData.cFileName) != 0 && strcmp(BOGUS_FILE_2, FindFileData.cFileName) != 0) {   /* Not "." or ".." */
                if (table_ptr == num_threads) {
#pragma omp parallel num_threads(num_threads)
                    {
#if _MSC_VER >= 1400  && !defined(_AMD64_)
                        int id = omp_get_thread_num();
                        deflate_st(file_table[id]);
#endif
                    }
                    for (table_ptr = 0; table_ptr < num_threads; table_ptr++) {
                        zfree(file_table[table_ptr]);
                    }
                    table_ptr = 0;
                }
                file_table[table_ptr] = (char*)zalloc(strlen(FindFileData.cFileName) + strlen(dirname) + 2);
                strcpy(file_table[table_ptr], dirname);
                strcat(file_table[table_ptr], "\\");
                strcat(file_table[table_ptr], FindFileData.cFileName);
                table_ptr++;                    /* One else entry is done */
            }
            if (FindNextFile(FindHandle, &FindFileData) != TRUE) {
                if (table_ptr > 0) {
#pragma omp parallel num_threads(table_ptr)
                    {
#if _MSC_VER >= 1400  && !defined(_AMD64_)
                        int id = omp_get_thread_num();
                        deflate_st(file_table[id]);
#endif
                    }
                }
                break;
            }
        }
        /* Free all data */
        for (table_ptr = 0; table_ptr < num_threads; table_ptr++) {
            zfree(file_table[table_ptr]);
        }
        zfree(file_table);
        zfree(valid_search_name);
        FindClose(FindHandle);
        return NULL;
    }
#endif
#else
    /* Non-Win32 version, i.e. Linux */
    DIR *dir_ptr;
    struct dirent *entry_ptr;
    char **file_table;
    int table_ptr, i;
#ifdef GZIP_PTHREADS
    char *dirname;
    pthread_t *pthreads;
    pthread_attr_t attr;
#endif

#ifdef GZIP_PTHREADS
    dirname = (char*)param;
#endif
    if ((dir_ptr = opendir(dirname)) == NULL) {
        perror("opendir failure");
        return 0;
    }
    if (num_threads == 1) {         /* Explicitly single-thread version */
        while ((entry_ptr = readdir(dir_ptr)) != NULL) {
            char *full_filename;
            size_t full_name_size;
            if (strcmp(BOGUS_FILE_1, entry_ptr->d_name) == 0 || strcmp(BOGUS_FILE_2, entry_ptr->d_name) == 0)
                continue;               /* Don't process "." and ".." */
            full_name_size = strlen(dirname);
            full_filename = (char*)zalloc(full_name_size+strlen(entry_ptr->d_name)+2);
            strcpy(full_filename, dirname);
            strcat(full_filename, "/");
            strcat(full_filename, entry_ptr->d_name);
            deflate_st(full_filename);
            zfree(full_filename);
        }
    } else {
        file_table = (char**)zalloc(num_threads * sizeof(char*));
        table_ptr = 0;
#ifdef GZIP_PTHREADS
        pthreads = (pthread_t*)zalloc(num_threads * sizeof(pthread_t));
        pthread_attr_init(&attr);
#endif
        while(1) {
            size_t str_size;

            if ((entry_ptr = readdir(dir_ptr)) == NULL) {       /* End of directory */
                if (table_ptr > 0) {            /* Some of files were not processed */
#ifdef GZIP_PTHREADS
                    for(i = 0; i<table_ptr; i++)
                        pthread_create(&pthreads[i], &attr, deflate_st, (void*)file_table[i]);
                    for(i = 0; i<table_ptr; i++)
                        pthread_join(pthreads[i], NULL);
#else
#pragma omp parallel num_threads(table_ptr)
                    {
                        int id = omp_get_thread_num();
                        deflate_st(file_table[id]);
                    }
                    for (i = 0; i < table_ptr; i++) {
                        zfree(file_table[i]);
                    }
#endif
                }
                break;              /* Leave while(1) */
            }
            if (strcmp(BOGUS_FILE_1, entry_ptr->d_name) == 0 || strcmp(BOGUS_FILE_2, entry_ptr->d_name) == 0)
                continue;                   /* Go to next file if "." or ".." */
            if (table_ptr == num_threads) {     /* All threads are ready to go */
#ifdef GZIP_PTHREADS
                for (table_ptr=0; table_ptr<num_threads; table_ptr++) {
                    pthread_create(&pthreads[table_ptr], &attr, deflate_st, (void*)file_table[table_ptr]);
                }
                for (table_ptr=0; table_ptr<num_threads; table_ptr++) {
                    pthread_join(pthreads[table_ptr], NULL);
                }
#else           /* OpenMP version */
#pragma omp parallel num_threads(num_threads)
                {
                    int id = omp_get_thread_num();
                    deflate_st(file_table[id]);
                }
#endif
                for (table_ptr = 0; table_ptr < num_threads; table_ptr++) {
                    zfree(file_table[table_ptr]);
                }
                table_ptr = 0;
            }
            str_size = strlen(entry_ptr->d_name);
            file_table[table_ptr] = (char*)zalloc(str_size + strlen(dirname) + 2);
            strcpy(file_table[table_ptr], dirname);
            strcat(file_table[table_ptr], "/");
            strcat(file_table[table_ptr], entry_ptr->d_name);
            table_ptr++;
        }   /* while (1) */
        zfree(file_table);
#ifdef GZIP_PTHREADS
        zfree(pthreads);
        pthread_attr_destroy(&attr);
#endif
    }   /* num_threads */
    closedir(dir_ptr);
#endif
    return NULL;
}

