/*
//
//                  INTEL CORPORATION PROPRIETARY INFORMATION
//     This software is supplied under the terms of a license agreement or
//     nondisclosure agreement with Intel Corporation and may not be copied
//     or disclosed except in accordance with the terms of that agreement.
//          Copyright(c) 2007-2008 Intel Corporation. All Rights Reserved.
//
//     Intel(R) Integrated Performance Primitives Data Compression Domain-based GZIP
//
//  By downloading and installing this sample, you hereby agree that the
//  accompanying Materials are being provided to you under the terms and
//  conditions of the End User License Agreement for the Intel(R) Integrated
//  Performance Primitives product previously accepted by you. Please refer
//  to the file ippEULA.rtf located in the root directory of your Intel(R) IPP
//  product installation for more information.
//
*/
#if defined(GZIP_PTHREADS) || (defined(_OPENMP) && defined(GZIP_PARALLEL))

#include <stdio.h>
#include <sys/stat.h>
#include <ippcore.h>
#include <ippdc.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#ifdef _WIN32
    #include <windows.h>
    #include <io.h>
#else
    #include <unistd.h>
#endif
#ifdef GZIP_PTHREADS
    #include <pthread.h>
#else
    #include <omp.h>
#endif

#include "ipp_gzip.h"

#define stop_deflating_mt() { zfree(offsets); zfree(crc_array); zfree(real_comment); }

#ifdef GZIP_PTHREADS
struct cs_params {
    char *filename;
    int my_chunk;
    int max_chunk;
    size_t chunk_size;
    GZIP_FILE *tmp_fd;
};
struct cc_params {
    char *filename;
    int max_chunk;
    size_t chunk_size;
    Ipp32u *crc;
    Ipp32u *crc_array;
};
static void *compress_n_save(void*);
static void *compute_crc(void*);
#else       /* OMP version*/
static int compress_n_save(char *, int, int, size_t, GZIP_FILE*);
static int compute_crc(char*, int, size_t, Ipp32u*, Ipp32u*);
#endif


/**
 * Multi-threaded version of "deflate" function
 * @param filename - name of file to compress
 */
void *deflate_mt(char *filename)
{
    struct stat file_info;
    unsigned int chunks; /* Number of chunks to slice file to */
    size_t chunk_size;
    size_t comment_size;
    char out_filename[MAX_NAME];
    char base_name[MAX_NAME];
    UBYTE *zip_header;      /* buffer to store zip header */
    UBYTE *zip_hdr_ptr;
    long comment_offset;            /* Offset from beginning of file to comment field */
    unsigned int i;
    int status_write;
    UBYTE tail_buf[8];
    UBYTE *tail_ptr;
    UBYTE *ptr;
    Ipp32u comment_crc;
    UBYTE comment_crc_xored;
    UBYTE *inbuf;
    size_t incnt;
    size_t *offsets;       /* Binary offset array */
    Ipp32u *crc_array;   /* CRC value array */
    UBYTE *real_comment;  /* Pointer to comment buffer */
    GZIP_FILE *fd_out;          /* File descriptor for output */
    Ipp32u crc;          /* Accumulated CRC */
    GZIP_FILE **tmp_fd_vector;  /* Vector of temporary file descriptors */
    short reg_file_output;
    GZIP_FILE* real_output;     /* File variable for writing to device or pipe */
    Ipp64u stat_start_clocks, stat_finish_clocks;
#ifdef GZIP_PTHREADS
    struct cs_params *pcompress_params;
    struct cc_params compute_crc_params;
    pthread_t *pthreads;
    pthread_attr_t attr;
#else
    int team_threads;
#endif

    if (from_stdin) {                       /* I.e. deflate from stdin */
        if (ISATTY(FILENO(stdin))) {
            warn_message("don't read non-compressed data from terminal");
            exit(1);
        }
        if (!is_file(FILENO(stdin)))
            return deflate_st(NULL);        /* If stdin is not a regular file (i.e. we cannot seek on it) do compression sequentially */
    }
    /* If stdout is not a regular file
        - we will use temporary file to collect ALL compressed data
        - will copy temporary file to stdout
        If stdout is a file
        - we will use it as "chunk 0" output file, thus saving time on I/O
    */
    if (opt_to_stdout) {
        if (ISATTY(FILENO(stdout))) {
            warn_message("don't write output data to terminal");
            exit(1);
        }
        reg_file_output = (short)is_file(FILENO(stdout));  /* Just marker - regular/not regular */
    }
    else
        reg_file_output = 1;
    if (from_stdin) {
        filename = "stdin";
    } else {
        if (known_suffix(filename))
            return NULL;
#if _MSC_VER >= 1400  && !defined(_AMD64_)
        if (_stat(filename, &file_info) == -1) {
#else
        if (stat(filename, &file_info) == -1) {
#endif
            error_file(filename);
            return NULL;
        }
        /* Look if file is directory */
#if _MSC_VER >= 1400  && !defined(_AMD64_)
        if ((file_info.st_mode & _S_IFDIR) != 0) {
#else
        if ((file_info.st_mode & S_IFDIR) != 0) {
#endif
            return compress_dir_st(filename);
        }
        if (opt_dbg_slice_size && file_info.st_size < opt_slice_size)       /* "-j minsize" option found */
            return deflate_st(filename);
        else if (opt_dbg_slice_size == 0 && file_info.st_size < MIN_LENGTH_TO_SLICE) { /* If file is too small, process it with no chunks */
            return deflate_st(filename);
        }
    }
    chunks = get_cpu_count();       /* Each chunk in it's own thread */
    if (chunks == 1) {              /* Single thread environment */
        return deflate_st(filename);
    }
    if (opt_perf_stat)
        stat_start_clocks = ippGetCpuClocks();
    chunk_size = file_info.st_size/chunks;
    comment_size = (chunks-1)*16 + 4;   /* Size of comment field */
    real_comment = (UBYTE*)zalloc(comment_size);
    offsets = (size_t*)zalloc((chunks-1)*sizeof(size_t));   /* We need N-1 offsets */
    crc_array = (Ipp32u*)zalloc((chunks-1)*sizeof(Ipp32u));
    create_output_name(filename, out_filename);
    if (opt_to_stdout) {
        if ((fd_out = zopen_stdout()) == NULL) {
            error_file("stdout");
            stop_deflating_mt();
            return NULL;
        }
    } else {
        if ((fd_out = out_open(out_filename, file_info.st_size * 2)) == NULL) {
            error_file(out_filename);
            stop_deflating_mt();
            return NULL;
        }
    }
    fd_out->uncompressed_size = file_info.st_size * 2;
    zip_header = (UBYTE*)zalloc(GZIP_HEADER_SIZE + strlen(filename) + 1 + comment_size);
    zip_hdr_ptr = zip_header;
    if (!opt_noname) 
        get_base_name(filename, base_name);
    zip_hdr_ptr = create_gzip_hdr(zip_hdr_ptr, base_name, &file_info, 1);           /* Create gzip header structure with comment */

    comment_offset = (long)(zip_hdr_ptr - zip_header);

    /* Fake comment field for now */
    for (i=0; i<comment_size-1; i++) {
        PUT_BYTE(zip_hdr_ptr, 'U');
    }
    PUT_BYTE(zip_hdr_ptr, 0);
    if (reg_file_output) {          /* If output is a file, put GZIP header. If not, keep gzip header, it will be written when real output is done */
        /* Write header to file */
        status_write = (int)zwrite(fd_out, zip_header, zip_hdr_ptr - zip_header);
        zfree(zip_header);
        if (status_write == -1) {
            error_file(out_filename);
            stop_deflating_mt();
            zclose(fd_out);
            return 0;
        }
    }
    /* Modify thread number to +1 to use 1 additional thread for CRC computing */
    tmp_fd_vector = (GZIP_FILE**)zalloc(sizeof(GZIP_FILE*) * chunks);
    for (i=0; i<chunks; i++) {
        char tmp_name[MAX_NAME];
        if (i > 0) {   /* Thread #1 will be used fd_out as compressed data output */
            if ((tmp_fd_vector[i] = ZOPEN(mk_chunk_name(base_name, i, tmp_name), GZIP_FILE_WRITE, (file_info.st_size/chunks)*2)) == NULL) {
                perror("cannot create temporary file");
                abort();
            }
        } else {
            tmp_fd_vector[i] = fd_out;
        }
    }
#ifndef GZIP_PTHREADS       /* OMP version */
    team_threads = chunks+1;
#ifdef __INTEL_COMPILER
    kmp_set_library_throughput();
    kmp_set_blocktime(0);
#endif
#pragma omp parallel num_threads(team_threads)
    {
        int id = omp_get_thread_num();
        if (0 == id) {
            /* Start CRC computing */
            compute_crc(filename, chunks-1, chunk_size, &crc, crc_array);
        } else {
            /* Start chunk compression */
            compress_n_save(filename, id-1, chunks-1, chunk_size, tmp_fd_vector[id-1]);
        }
    }
#else                   /* PTHREAD version */
    pcompress_params = (struct cs_params*)zalloc(chunks * sizeof(struct cs_params));
    pthreads = (pthread_t*)zalloc((chunks+1) * sizeof(pthread_t));
    pthread_attr_init(&attr);
    compute_crc_params.filename = filename;
    compute_crc_params.max_chunk = chunks-1;
    compute_crc_params.chunk_size = chunk_size;
    compute_crc_params.crc = &crc;
    compute_crc_params.crc_array = crc_array;
    pthread_create(&pthreads[0], &attr, compute_crc, &compute_crc_params);
    for (i=1; i<chunks+1; i++) {
        pcompress_params[i-1].filename = filename;
        pcompress_params[i-1].my_chunk = i-1;
        pcompress_params[i-1].max_chunk = chunks-1;
        pcompress_params[i-1].chunk_size = chunk_size;
        pcompress_params[i-1].tmp_fd = tmp_fd_vector[i-1];
        pthread_create(&pthreads[i], &attr, compress_n_save, &pcompress_params[i-1]);
    }
    for (i=0; i<chunks+1; i++) {
        pthread_join(pthreads[i], NULL);
    }
#endif
    /* Build offset table */
    if (reg_file_output)
        offsets[0] = ztell(fd_out);
    else
        offsets[0] = ztell(fd_out) + (zip_hdr_ptr - zip_header);    /* Remember, that we didn't put out the header */
    for (i=1; i<chunks-1; i++) {
        offsets[i] = offsets[i-1] + ztell(tmp_fd_vector[i]);
    }
    inbuf = (UBYTE*) zalloc(INBUF_SIZE);
    /* Prepare offset/crc arrays */
    ptr = real_comment;
    *ptr++ = COMMENT_MAGIC_HEADER;
    ptr += sprintf((char*)ptr, "%1X", chunks);
    for (i=0; i<chunks-1; i++) {
        ptr += sprintf((char*)ptr, "%08lX", offsets[i]);
    }
    for (i=0; i<chunks-1; i++) {
        ptr += sprintf((char*)ptr, "%08X", crc_array[i]);
    }
    comment_crc = 0;
    ippsCRC32_8u(real_comment, (int)(ptr - real_comment), &comment_crc);
    comment_crc_xored = (UBYTE)((comment_crc & 0xFF) ^ ((comment_crc >> 8) & 0xFF) ^ ((comment_crc >> 16) & 0xFF) ^ ((comment_crc >> 24) & 0xFF));
    if (comment_crc_xored == 0) 
        comment_crc_xored = 1;              /* Text field ((*/
    *ptr++ = comment_crc_xored;
    real_comment[comment_size-1] = '\0';
    if (!reg_file_output) {                 /* Sending output to device or pipe */
        real_output = zopen_stdout();
        /* Write GZIP header and comment field */
        memcpy(zip_header + comment_offset, real_comment, comment_size);
        zwrite(real_output, zip_header, zip_hdr_ptr - zip_header);
        /* Copy fd_out (chunk 0) to real output file */
        zseek(fd_out, 0, SEEK_SET);
        while (!zeof(fd_out)) {
            incnt = zread(fd_out, inbuf, INBUF_SIZE);
            zwrite(real_output, inbuf, incnt);
        }
        zclose(fd_out);
        UNLINK(out_filename);                   /* We don't need it anymore */
        fd_out = real_output;                   /* Will use below code for fd_out */
        zfree(zip_header);
    }
    for (i=1; i<chunks; i++) {
#ifdef _WIN32
        if (opt_io_read_method == METHOD_LOW && opt_io_write_method == METHOD_MMAP) {   /* Tmp file used mmap for write, 
                                                            but low-level functions must be used for read */
            tmp_fd_vector[i]->fd = _open_osfhandle((intptr_t)tmp_fd_vector[i]->hFile, _O_RDONLY);
            if (tmp_fd_vector[i]->fd == -1) {
                perror("Cannot reopen tmp file for reading");
                exit(1);
            }
            tmp_fd_vector[i]->mode = GZIP_FILE_READ;
        }
#endif
        if (zseek(tmp_fd_vector[i], 0, SEEK_SET)) { 
            perror("cannot rewind temporary file"); 
            abort(); 
        }
        while (!zeof(tmp_fd_vector[i])) {
            if ((incnt = zread(tmp_fd_vector[i], inbuf, INBUF_SIZE)) == -1) {
                perror("Cannot read from tmp file");
                abort();
            }
            zwrite(fd_out, inbuf, incnt);
        }
        zclose(tmp_fd_vector[i]);
        remove_tmp_file(base_name, i);
    }
    /* Finish file with CRC and total file length */
    tail_ptr = tail_buf;
    PUT_LONG(tail_ptr, crc);
    PUT_LONG(tail_ptr, (unsigned long)file_info.st_size);
    zwrite(fd_out, tail_buf, 8);
    if (reg_file_output) {
        /* Finish the comment */
        if (zseek(fd_out, comment_offset, 0) == (-1L))
            perror("final seek failed");
        else
            zwrite(fd_out, real_comment, comment_size);
    }
    zfree(inbuf);
    zfree(real_comment);
    zfree(offsets);
    zfree(crc_array);
    zfree(tmp_fd_vector);
#ifdef GZIP_PTHREADS
    zfree(pcompress_params);
#endif
    if (!opt_to_stdout && filename != NULL) {
        /* Print out compression statistics */
        if (opt_perf_stat) {
            stat_finish_clocks = ippGetCpuClocks();
            fprintf(stderr, "%s:\t deflating -- ", filename);
            print_io_methods();
            fprintf(stderr, " MT -- %6.1f clocks per input symbol\n", (float)(stat_finish_clocks - stat_start_clocks) / file_info.st_size);
        }
        if (opt_verbose) {
            fprintf(stderr, "%s:\t %6.1f%% -- replaced with %s\n", filename, 100*(1.-(float)fd_out->bytes_out/file_info.st_size), out_filename);
        }
        if (UNLINK(filename) != 0)
            perror("unlink");
    }
    zclose(fd_out);
    return NULL;
}

#ifdef GZIP_PTHREADS
static void *compress_n_save(void *params)
#else
static int compress_n_save(char *filename, int my_chunk, int max_chunk, size_t chunk_size, GZIP_FILE *tmp_fd)
#endif
{
    GZIP_FILE *fd;
    UBYTE *in_ptr, *out_ptr;
    size_t incnt;
    IppStatus status;
    IppLZ77State_8u *ippState;
    IppLZ77Flush flush_mode;
    IppLZ77Pair *c_pair, *p_pair;
    int c_index, c_len, len, in_len;
    size_t out_len;
    int status_write;
    size_t buff_size;
    size_t bytes_to_read;
    size_t read_size;
    UBYTE *inbuf;
    UBYTE *outbuf;
    IppLZ77ComprLevel compr_level;
    IppLZ77HuffMode huff_mode;
    UBYTE stored_block_save[STORED_BUF_SIZE];
    UBYTE *stored_block_ptr;
    size_t stored_block_idx_low, stored_block_idx_high;
    size_t stored_block_watermark, stored_block_len; 
    int can_use_stored;
    /* DEBUG */
    size_t len_add, len_before;
#ifdef GZIP_PTHREADS
    struct cs_params *param_ptr;
    char *filename;
    int my_chunk;
    int max_chunk;
    size_t chunk_size;
    GZIP_FILE *tmp_fd;
#endif

    inbuf = (UBYTE*)zalloc(INBUF_SIZE);
    outbuf = (UBYTE*)zalloc(OUTBUF_SIZE);
#ifdef GZIP_PTHREADS
    param_ptr = (struct cs_params*)params;
    filename = param_ptr->filename;
    my_chunk = param_ptr->my_chunk;
    max_chunk = param_ptr->max_chunk;
    chunk_size = param_ptr->chunk_size;
    tmp_fd = param_ptr->tmp_fd;
#endif
    buff_size = INBUF_SIZE;
    bytes_to_read = chunk_size;
    if ((fd = ZOPEN(filename, GZIP_FILE_READ, 0)) == NULL) {
        error_file(filename);
        zfree(inbuf);
        zfree(outbuf);
        return 0;
    }
    if (my_chunk > 0) {
        if (zseek(fd, (long)(my_chunk * chunk_size), 0)) {
            perror("cannot seek");
            zclose(fd);
            zfree(inbuf);
            zfree(outbuf);
            return 0;
        }
    }
/* Init IPP structure */
    if (opt_compress_fast) {
        compr_level = IppLZ77FastCompr;
    } else if (opt_compress_best) {
        compr_level = IppLZ77BestCompr;
    } else
        compr_level = IppLZ77AverageCompr;
    status = ippsEncodeLZ77InitAlloc_8u(compr_level, IppLZ77CRC32, &ippState);
    ippsEncodeLZ77SetStatus_8u(IppLZ77StatusLZ77Process, ippState);
    ippsEncodeLZ77GetPairs_8u(&c_pair, &c_index, &c_len, ippState);
    p_pair = c_pair + c_index;
    len = c_len;
    out_ptr = outbuf;
    out_len = OUTBUF_SIZE;
    stored_block_watermark = 0;
    can_use_stored = 1;
    while (bytes_to_read && !zeof(fd)) {
        read_size = (bytes_to_read < buff_size && my_chunk != max_chunk)? bytes_to_read : buff_size;
        incnt = zread(fd, inbuf, read_size);
        bytes_to_read -= incnt;
        if (zeof(fd) || (bytes_to_read == 0)) {
            if (my_chunk == max_chunk) {
            flush_mode = IppLZ77FinishFlush;
                bytes_to_read = 0;
        } else {
            flush_mode = IppLZ77SyncFlush;
        }
        } else {
            flush_mode = IppLZ77NoFlush;
        }
        in_ptr = inbuf;
        in_len = (int)incnt;
        while(in_len > 0) {
            stored_block_idx_low = in_ptr - inbuf;       /* Save low index value for possible stored block ops */
            ippsEncodeLZ77SetStatus_8u(IppLZ77StatusLZ77Process, ippState);
            status = ippsEncodeLZ77_8u(&in_ptr, &in_len, &p_pair, &len, flush_mode, ippState);
            stored_block_idx_high = in_ptr - inbuf;
            stored_block_len = stored_block_idx_high - stored_block_idx_low;
            if (can_use_stored && (stored_block_watermark+stored_block_len) < STORED_BUF_SIZE) {
                memcpy(&stored_block_save[stored_block_watermark], &(inbuf[stored_block_idx_low]), stored_block_len);
                stored_block_watermark += stored_block_len;
                can_use_stored = 1;
            } else {    /* Processed block is too long to be stored */
                stored_block_watermark = 0;
                can_use_stored = 0;
            }
            switch (status) {
            case ippStsNoErr:
                if (bytes_to_read > 0) {
                    break;                      /* Goto reading of next buffer */
                }
            case ippStsDstSizeLessExpected:
                p_pair = c_pair;
                len = c_len - len;
                status = ippsEncodeLZ77SetStatus_8u(IppLZ77StatusHuffProcess, ippState);
                stored_block_ptr = stored_block_save;
                len_add = 0;
                while(1) {
                    if (!opt_force_dynamic)
                        status = ippsEncodeLZ77SelectHuffMode_8u(p_pair, len, &huff_mode, ippState);
                    else
                        huff_mode = IppLZ77UseDynamic;
                    switch (huff_mode) {
                    case IppLZ77UseFixed:
                        len_before = out_len;
                        status = ippsEncodeLZ77FixedHuff_8u(&p_pair, &len, &out_ptr, (int*)&out_len, (in_len > 0)? IppLZ77NoFlush : flush_mode, ippState);
                        if (status == ippStsNoErr || status == ippStsStreamEnd) {
                            if (opt_verbose) {
                            }
                            len_add = 0;
                        } else {
                            len_add += len_before-out_len;
                        }
                        break;
                    case IppLZ77UseStored:
                        if (can_use_stored) {
                            len_before = stored_block_watermark;
                            status = ippsEncodeLZ77StoredBlock_8u(&stored_block_ptr, (int*)&stored_block_watermark, &out_ptr, (int*)&out_len, 
                                (in_len > 0)? IppLZ77NoFlush : flush_mode, ippState);
                            if (status == ippStsNoErr) {
                                if (opt_verbose) {
                                }
                                len_add = 0;
                            } else {
                                len_add += len_before-stored_block_watermark;
                            }
                            break;
                        }
                    case IppLZ77UseDynamic:
                        len_before = out_len;
                        status = ippsEncodeLZ77DynamicHuff_8u(&p_pair, &len, &out_ptr, (int*)&out_len, 
                            (in_len > 0 && flush_mode == IppLZ77FinishFlush)? IppLZ77SyncFlush : flush_mode, ippState);
                        if (status == ippStsNoErr) {
                            if (opt_verbose) {
                            }
                            len_add = 0;
                        } else {
                            len_add += len_before-out_len;
                        }
                        break;
                    }
                    if (status == ippStsDstSizeLessExpected || status == ippStsStreamEnd || (status == ippStsNoErr && out_len == 0)) {
                        status_write = (int)zwrite(tmp_fd, outbuf, OUTBUF_SIZE-out_len);
                        if (status_write == -1) { 
                            perror("cannot write to temporary file"); 
                            abort(); 
                        }
                        out_ptr = outbuf;
                        out_len = OUTBUF_SIZE;
                    }
                    if (status == ippStsNoErr || status == ippStsStreamEnd)
                        break;
                }
                p_pair = c_pair;
                len = c_len;
                can_use_stored = 1;
                stored_block_watermark = 0;
                break;
            default:
                warn_message("Fatal IPP error!");
                break;
            }
        }
    }
/* End of input buffer processing */
    if (out_len < OUTBUF_SIZE) {
        status_write = (int)zwrite(tmp_fd, outbuf, OUTBUF_SIZE-out_len);
        if (status_write == -1) { 
            perror("cannot write to temporary file"); 
            abort(); 
        }
    }
    ippsLZ77Free_8u(ippState);
    if (my_chunk > 0)
        ztruncate(tmp_fd);
#ifdef GZIP_PTHREADS
    return NULL;
#else
    zclose(fd);
    zfree(inbuf);
    zfree(outbuf);
    return 1;
#endif
}
/*! \brief Size of CRC computing buffer size */
#define CRC_BUFSIZE (64*1024)
/**
 * Computes input file CRC in separated thread
 * @param filename  - input file name
 * @param max_chunk - max number of chunks to split
 * @param chunk_size - size of each chunk
 */
#ifdef GZIP_PTHREADS
static void *compute_crc(void *params)
#else
static int compute_crc(char *filename, int max_chunk, size_t chunk_size, Ipp32u *crc, Ipp32u *crc_array)
#endif
{
    UBYTE buffer[CRC_BUFSIZE];
    size_t incnt;
    GZIP_FILE *fd;
    size_t bytes_for_crc;
    size_t bytes_to_read;
    int chunk_no;
    size_t total_bytes;
#ifdef GZIP_PTHREADS
    struct cc_params *param_ptr;
    char *filename;
    int max_chunk;
    size_t chunk_size;
    Ipp32u *crc;
    Ipp32u *crc_array;
#endif

#ifdef GZIP_PTHREADS
    param_ptr = (struct cc_params*)params;
    filename = param_ptr->filename;
    max_chunk = param_ptr->max_chunk;
    chunk_size = param_ptr->chunk_size;
    crc = param_ptr->crc;
    crc_array = param_ptr->crc_array;
#endif
    if ((fd = ZOPEN(filename, GZIP_FILE_READ, 0)) == NULL) {
        perror("cannot open file for CRC computing");
        return 0;
    }
    chunk_no = 0;
    bytes_for_crc = chunk_size;
    total_bytes = 0;
    *crc = 0;
    while (!zeof(fd)) {
        bytes_to_read = (bytes_for_crc < sizeof(buffer))? bytes_for_crc : sizeof(buffer);
        incnt = zread(fd, buffer, bytes_to_read);
        bytes_for_crc -= incnt;
        total_bytes += incnt;
        ippsCRC32_8u(buffer, (int)incnt, crc);
        if (chunk_no < max_chunk) {
            if (bytes_for_crc == 0) {
                crc_array[chunk_no++] = *crc;
                bytes_for_crc = chunk_size;
            }
        } else {
            if (bytes_for_crc == 0) {
                if (!zeof(fd)) {
                    bytes_for_crc = chunk_size;
                    continue;
                }
            }
        }
    }
    zclose(fd);
#ifdef GZIP_PTHREADS
    return NULL;
#else
    return 1;
#endif
}

#endif /* OPENMP && GZIP_PARALLEL */
