/*
//
//                  INTEL CORPORATION PROPRIETARY INFORMATION
//     This software is supplied under the terms of a license agreement or
//     nondisclosure agreement with Intel Corporation and may not be copied
//     or disclosed except in accordance with the terms of that agreement.
//          Copyright(c) 2007-2008 Intel Corporation. All Rights Reserved.
//
//     Intel(R) Integrated Performance Primitives Data Compression Domain-based GZIP
//
//  By downloading and installing this sample, you hereby agree that the
//  accompanying Materials are being provided to you under the terms and
//  conditions of the End User License Agreement for the Intel(R) Integrated
//  Performance Primitives product previously accepted by you. Please refer
//  to the file ippEULA.rtf located in the root directory of your Intel(R) IPP
//  product installation for more information.
//
*/
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#ifndef _WIN32
    #include <unistd.h>
#endif

#include <ippcore.h>
#include <ippdc.h>

#include "ipp_gzip.h"

struct inflate_data_pack {
    UBYTE *inbuf;
    UBYTE *outbuf;
    UBYTE *inptr;
    UBYTE *outptr;
    IppLZ77State_8u *ippState;
    IppLZ77Pair *c_pair;
    IppLZ77Pair *p_pair;
};

static void stop_inflating(struct inflate_data_pack *pdata)
{
    zfree(pdata->inbuf);
    zfree(pdata->outbuf);
    ippsLZ77Free_8u(pdata->ippState);
    zfree(pdata);
}

static size_t fill_inbuf(struct inflate_data_pack *pdata, GZIP_FILE *fd_in)
{
    return zread(fd_in, pdata->inbuf, OUTBUF_SIZE);
}

/**
 * Decompress single file in single-thread environment
 * @param filename
 */
void inflate_st(char *filename)
{
    GZIP_FILE *fd_in;
    GZIP_FILE *fd_out;
    size_t incnt;
    UBYTE flag_byte, method;
    long timestamp;
    char *orig_fname;
    UBYTE *comment_field;
    UBYTE *ptr;
    char *basename;         /* Base name (no path, no extension) of input file */
    struct inflate_data_pack *pdata;
    IppStatus status;
    UBYTE *inptr, *outptr;
    size_t inlen;
    size_t outlen;
    IppLZ77Flush flush_mode;
    IppLZ77HuffMode huff_mode;
    size_t outcnt, total_out;
    Ipp32u crc, stored_crc;
    size_t insize, outsize;
    int error_found;
    Ipp64u stat_start_clocks, stat_finish_clocks;
    /* DEBUG */
    size_t len_add, len_before;

    total_out = 0L;
    crc = 0L;
    /* For decompression output buffer should be longer than input buffer */
    insize = OUTBUF_SIZE;
    outsize = INBUF_SIZE;
    /* Allocate data pack structure */
    pdata = (struct inflate_data_pack*) zalloc(sizeof(struct inflate_data_pack));
    /* Open input file */
    if (from_stdin) {      /* Inflating stdin */
        fd_in = zopen_stdin();
    } else {
        if ((fd_in = ZOPEN(filename, GZIP_FILE_READ, 0)) == NULL) {
            error_file(filename);
            return;
        }
    }
    /* Allocate input/output buffers */
    pdata->inbuf = (UBYTE*) zalloc(insize);
    /* Read 1-st block of file */
    incnt = zread(fd_in, pdata->inbuf, insize);
    if (incnt == -1) {
        error_file(filename); 
        zclose(fd_in); 
        zfree(pdata->inbuf); 
        zfree(pdata); 
        return;
    }
    /* Check gzip header */
    if (pdata->inbuf[0] != GZIP_HDR_1 || pdata->inbuf[1] != GZIP_HDR_2) {
        warn_message("not in gzip format"); 
        zclose(fd_in); 
        zfree(pdata->inbuf); 
        zfree(pdata->outbuf); 
        zfree(pdata); 
        return;
    }
    method = pdata->inbuf[2];
    flag_byte = pdata->inbuf[3];
    if (method != DEFLATE) {
        fprintf(stderr, "%s: %s: unknown method %d -- not supported\n", prog_name, filename, method);
        zclose(fd_in); 
        zfree(pdata->inbuf); 
        zfree(pdata); 
        return;
    }
    /* GZIP header is verified. Look if we want to show list only */
    /* Get time stamp and original filename if exists */
    inptr = &pdata->inbuf[4];
    timestamp = GET_LONG(inptr); inptr += 4;
    if (opt_list) {
        show_list(filename, fd_in, timestamp);
        zclose(fd_in); 
        zfree(pdata->inbuf); 
        zfree(pdata->outbuf); 
        zfree(pdata); 
        return;
    }
    GET_BYTE(inptr);            /* Skip extra flags byte */
    GET_BYTE(inptr);            /* Skip OS type byte */
    if ((flag_byte & FNAME) && *inptr) {        /* If gzip flag says, that original filename exists */
        orig_fname = (char*)zalloc(strlen((const char*)inptr)+1);
        ptr = (UBYTE *)orig_fname;
        while (*ptr++ = GET_BYTE(inptr)) /* Copy original filename */
            ;
    } else {
        orig_fname = NULL;
    }
    /* Look for comment field */
    if ((flag_byte & FCOMMENT) && *inptr) {
        comment_field = (UBYTE *)zalloc(strlen((const char*)inptr)+1);
        ptr = comment_field;
        while (*ptr++ = GET_BYTE(inptr)) {  /* Copy comment */
            ;
        }
    }
    if (opt_perf_stat)
        stat_start_clocks = ippGetCpuClocks();
    /* Create output file */
    if (!opt_to_stdout) {
        if (!opt_name || (flag_byte & FNAME)==0) {       /* Use don't want to restore original name */
            basename = (char*)zalloc(strlen((const char*)filename)+1);
            if (!get_nosuf_name(filename, basename)) {
                zfree(pdata->inbuf);
                zfree(pdata);
                zclose(fd_in);
                return;
            }
        } else {
            size_t str_size;
            str_size = strlen((const char*)filename);
            basename = (char*)zalloc(str_size+strlen(orig_fname)+1);
            restore_output_name(filename, orig_fname, basename);
        }
    }
    if (!opt_test) {
        if (!opt_to_stdout) {
            if ((fd_out = out_open(basename, zsizeuncompressed(fd_in))) == NULL) {
                error_file(basename); 
                zfree(pdata->inbuf); 
                zfree(pdata->outbuf); 
                zfree(pdata); 
                zclose(fd_in); 
                return;
            }
            fd_out->uncompressed_size = zsizeuncompressed(fd_in);
        } else
            fd_out = zopen_stdout();
    } else if (opt_verbose && filename != NULL) {
        fprintf(stderr, "%s:\t", filename);
    }
    /* Inflate */
    status = ippsDecodeLZ77InitAlloc_8u(IppLZ77CRC32, &pdata->ippState);
    status = ippsDecodeLZ77SetStatus_8u(IppLZ77InflateStatusInit, pdata->ippState);
    status = ippsDecodeLZ77SetStatus_8u(IppLZ77InflateStatusHuffProcess, pdata->ippState);
    /* Time to create output buffer */
    pdata->outbuf = (UBYTE *) zalloc(outsize);
    outptr = pdata->outbuf;
    outlen = outsize;
    inlen = incnt - (int)(inptr - pdata->inbuf);
    if (incnt < insize)
        flush_mode = IppLZ77FinishFlush;
    else 
        flush_mode = IppLZ77NoFlush;
    len_add = 0;
    while(1) {
        status = ippsDecodeLZ77GetBlockType_8u( &inptr, (int*)&inlen, &huff_mode, pdata->ippState);
        if (status == ippStsSrcSizeLessExpected) {
            inlen = fill_inbuf(pdata, fd_in);
            inptr = pdata->inbuf;
            if (inlen < insize)
                flush_mode = IppLZ77FinishFlush;
            status = ippsDecodeLZ77GetBlockType_8u( &inptr, (int*)&inlen, &huff_mode, pdata->ippState);
        }
        if(status != ippStsNoErr) {     /* DecodeGetBlockType cannot determine deflated block type */
            fprintf(stderr, "%s: %s: unknown method %d -- not supported\n", prog_name, orig_fname, 0);
            stop_inflating(pdata);
            zclose(fd_in);
            return;
        }
        len_before = inlen;
        if (huff_mode == IppLZ77UseFixed) {
            status = ippsDecodeLZ77FixedHuffFull_8u(&inptr, (int*)&inlen, &outptr, (int*)&outlen, flush_mode, pdata->ippState);
            if (status == ippStsNoErr || status == ippStsStreamEnd) {
                len_add = 0;
            } else {
                len_add += len_before-inlen;
            }
        }
        else if (huff_mode == IppLZ77UseDynamic) {
            status = ippsDecodeLZ77DynamicHuffFull_8u(&inptr, (int*)&inlen, &outptr, (int*)&outlen, flush_mode, pdata->ippState);
                len_add += len_before-inlen;
            if (status == ippStsNoErr || status == ippStsStreamEnd) {
                len_add = 0;
            }
        }
        else if (huff_mode == IppLZ77UseStored) {
            status = ippsDecodeLZ77StoredBlock_8u(&inptr, (int*)&inlen, &outptr, (int*)&outlen, pdata->ippState);
            if (status == ippStsNoErr || status == ippStsStreamEnd) {
                len_add = 0;
            } else {
                len_add += len_before-inlen;
            }
            /* Bug workaround */
            if (status == ippStsDstSizeLessExpected && inlen == 0 && outlen != 0)
                status = ippStsSrcSizeLessExpected;
        }
        if (status == ippStsSrcSizeLessExpected) {
            inlen = fill_inbuf(pdata, fd_in);
            inptr = pdata->inbuf;
            if (inlen < insize)
                flush_mode = IppLZ77FinishFlush;
        } else if (status == ippStsDstSizeLessExpected || status == ippStsStreamEnd) {
            outcnt = outsize - outlen;
            total_out += outcnt;
            ippsCRC32_8u(pdata->outbuf, (int)outcnt, &crc);
            if (!opt_test) 
                zwrite(fd_out, pdata->outbuf, outcnt);
            outptr = pdata->outbuf;
            outlen = outsize;
            if(status == ippStsStreamEnd) {
                break;
            }
        } else if (status != ippStsNoErr) {
            print_ipp_status(status);
            stop_inflating(pdata);
            zclose(fd_in);
            if (!opt_test) zclose(fd_out);
            return;
        }
    }
    /* Check CRC */
    stored_crc = GET_LONG(inptr); inptr += 4;
    error_found = 0;
    if (stored_crc != crc) {
        fprintf(stderr, "\n%s: %s: invalid compressed data--crc error\n", prog_name, filename);
    } 
    if (total_out != GET_LONG(inptr)) {  /* Check compressed file size */
        fprintf(stderr, "\n%s: %s: invalid compressed data--length error\n", prog_name, filename);
        error_found = 1;
    } 
    if (opt_test && opt_verbose && !error_found) {
        fprintf(stderr, " OK\n");
    }
    stop_inflating(pdata);
    /* Print decompression statistics */
    if (opt_perf_stat && !opt_test && !opt_to_stdout && filename != NULL) {
        stat_finish_clocks = ippGetCpuClocks();
        fprintf(stderr, "%s:\t inflating ", filename);
        print_io_methods();
        fprintf(stderr, " ST -- %6.1f clocks per input symbol\n", (float)(stat_finish_clocks - stat_start_clocks) / fd_in->bytes_in);
    }
    if (opt_verbose && !opt_test && !opt_to_stdout && filename != NULL) {
        fprintf(stderr, "%s:\t %6.1f%% -- replaced with %s\n", filename, 100*(1.-(float)fd_in->bytes_in/fd_out->bytes_out), basename);
    }
    zclose(fd_in);
    if (!opt_test) 
        zclose(fd_out);
    if (!opt_notime && !opt_to_stdout)          /* Need to restore original file creation time. */
        zcopytime(filename, basename, timestamp);
    if (!opt_to_stdout && filename != NULL && !opt_test) {
        if (UNLINK(filename) != 0)
            perror("cannot unlink");
    }
}


