/*
//
//                  INTEL CORPORATION PROPRIETARY INFORMATION
//     This software is supplied under the terms of a license agreement or
//     nondisclosure agreement with Intel Corporation and may not be copied
//     or disclosed except in accordance with the terms of that agreement.
//          Copyright(c) 2007-2008 Intel Corporation. All Rights Reserved.
//
//     Intel(R) Integrated Performance Primitives Data Compression Domain-based GZIP
//
//  By downloading and installing this sample, you hereby agree that the
//  accompanying Materials are being provided to you under the terms and
//  conditions of the End User License Agreement for the Intel(R) Integrated
//  Performance Primitives product previously accepted by you. Please refer
//  to the file ippEULA.rtf located in the root directory of your Intel(R) IPP
//  product installation for more information.
//
*/
#if defined(GZIP_PTHREADS) || (defined(_OPENMP) && defined(GZIP_PARALLEL))

#include <stdio.h>
#include <fcntl.h>
#ifdef _WIN32
    #include <sys/stat.h>
    #include <windows.h>
    #include <io.h>
#else
    #include <unistd.h>
#endif   
#include <string.h>
#include <stdlib.h>
#include <time.h>

#ifndef GZIP_PTHREADS
#include <omp.h>
#else
    #include <pthread.h>
#endif

#include <ippcore.h>
#include <ippdc.h>

#include "ipp_gzip.h"

/*! \brief Choose minimal of two values */
#define MINSIZE(a, b) ((a<b)? a: b)

static char OFFSET_FORMAT[] = "%08X";

#ifdef GZIP_PTHREADS
struct ic_params {
    int chunk;
    int max_chunk;
    Ipp32u from_crc;
    Ipp32u to_crc;
    long from_pos;
    long to_pos;
    size_t in_buffer_size;
    size_t out_buffer_size;
    char *filename;
    UBYTE *chunk0_inptr;
    UBYTE *chunk0_inbuf;
    GZIP_FILE *chunk0_fd_in;
    GZIP_FILE *tmp_fd;
    int *err_status;
    int targetFileSize;
};
static void *inflate_chunk(void*);
#else

static int inflate_chunk(int, Ipp32u, Ipp32u, unsigned long, unsigned long, size_t, size_t, char*, UBYTE*, UBYTE*, GZIP_FILE*, GZIP_FILE*, int*, int);

#endif

struct allocatable_data {
    GZIP_FILE *fd_in;
    UBYTE *inbuf;
    char *orig_fname;
    char *result_fname;
    long *offset_array;
    Ipp32u *crc_array;
    UBYTE *comment_field;

};

/**
 * Free all allocated storage
 * @param pdata
 */
static void free_data(struct allocatable_data *pdata)
{
    zfree(pdata->comment_field);
    zfree(pdata->crc_array);
    zfree(pdata->inbuf);
    zfree(pdata->offset_array);
    zfree(pdata->orig_fname);
    zfree(pdata->result_fname);
    zclose(pdata->fd_in);
}
/**
 * Multi-threaded version of inflate function
 * 
 * @param filename
 */
void inflate_mt(char *filename)
{
    struct allocatable_data pdata;
    size_t incnt;
    UBYTE flag_byte, method;
    UBYTE *inptr;
    UBYTE *ptr;
    Ipp32u comment_crc;
    size_t comment_field_length;
    UBYTE comment_crc_xored;
    int chunks;
    int i;
#if _MSC_VER >= 1400  && !defined(_AMD64_)
    struct _stat file_info;
#else
    struct stat file_info;
#endif
    long timestamp;
    size_t insize, outsize;
    GZIP_FILE *fd_out = NULL;
    int error_found;
    GZIP_FILE **tmp_fd_vector = NULL;          /* Array of temporary files to be used in low-level i/o */
    Ipp64u stat_start_clocks, stat_finish_clocks;
#ifdef GZIP_PTHREADS
    struct ic_params *p_params;
    pthread_t *pthreads;
    pthread_attr_t attr;
    int next_free_thread_slot = 0;
#else
    unsigned long from_pos, to_pos;
    Ipp32u from_crc, to_crc;
#endif
    DWORD targetFileSize;

    if (get_cpu_count() == 1) {
        inflate_st(filename);        /* No alternates */
        return;
    }
    if (from_stdin) {          /* I.e. inflate from stdin */
        if (ISATTY(FILENO(stdin))) {
            warn_message("don't read compressed data from terminal");
            exit(1);
        }
        inflate_st(NULL);
        return;
    }
    /* Get input file statistics */
#if _MSC_VER >= 1400  && !defined(_AMD64_)
    if (_stat(filename, &file_info) == -1) {
#else
    if (stat(filename, &file_info) == -1) {
#endif
        fprintf(stderr, "Cannot get file statistics\n");
        error_file(filename); return;
    }
    /* Allocate and initialize data structure for pointers */
    pdata.comment_field = NULL;
    pdata.crc_array = NULL;
    pdata.inbuf = NULL;
    pdata.offset_array = NULL;
    pdata.orig_fname = NULL;
    pdata.result_fname = NULL;
    /* Open input file */
    if ( (pdata.fd_in = ZOPEN(filename, GZIP_FILE_READ, 0)) == NULL ) {
        fprintf(stderr, "Cannot open input file for sharing\n");
        error_file(filename); return;
    }
    insize = OUTBUF_SIZE;
    outsize = INBUF_SIZE;
    pdata.inbuf = (UBYTE *) zalloc(insize);
    /* Read header, original filename, comment field */
    incnt = zread(pdata.fd_in, pdata.inbuf, insize);
    if ( incnt == -1 ) {
        fprintf(stderr, "Cannot read header\n");
        error_file(filename); 
        free_data(&pdata);
        return;
    }
    /* Check GZIP header */
    if ( pdata.inbuf[0] != GZIP_HDR_1 || pdata.inbuf[1] != GZIP_HDR_2 ) {
        warn_message("not in gzip format"); 
        free_data(&pdata);
        return;
    }
    method = pdata.inbuf[2];
    if ( method != DEFLATE ) {
        fprintf(stderr, "%s: %s: unknown method %d -- not supported\n", prog_name, filename, method);
        free_data(&pdata);
        return;
    }
    flag_byte = pdata.inbuf[3];
    /* Get time stamp and original filename if exists */
    inptr = pdata.inbuf + 4;
    timestamp = GET_LONG(inptr); inptr += 4;
    GET_BYTE(inptr);    /* Skip extra flags byte */
    GET_BYTE(inptr);    /* Skip OS flag byte */
    if ( (flag_byte & FNAME) && *inptr ) {        /* If gzip flag says, that original filename exists */
        pdata.orig_fname = (char*)zalloc(strlen((const char*)inptr)+1);
        ptr = (UBYTE *)pdata.orig_fname;
        while ( (*ptr++ = GET_BYTE(inptr)) ) /* Copy original filename */
            ;
    }
    if (opt_list) {         /* We want just to show archive content */
        show_list(filename, pdata.fd_in, timestamp);
        free_data(&pdata);
        return;
    }
    /* Look for comment field */
    if ( (flag_byte & FCOMMENT) && *inptr ) {
        pdata.comment_field = (UBYTE *)zalloc(strlen((const char*)inptr)+1);
        ptr = pdata.comment_field;
        while ( (*ptr++ = GET_BYTE(inptr)) ) /* Copy comment */
            ;
        }
    if ( (flag_byte & FCOMMENT) == 0                 /* If no comment exists */
         || (*(pdata.comment_field)) != COMMENT_MAGIC_HEADER ) {   /* or it is not multi-chunk comment */
        free_data(&pdata);
        inflate_st(filename);
        return;
    }
    /* Additionally check comment field with CRC */
    comment_field_length = strlen((const char*)pdata.comment_field);
    comment_crc = 0;
    ippsCRC32_8u(pdata.comment_field, (int)(comment_field_length-1), &comment_crc);
    comment_crc_xored = (UBYTE)((comment_crc & 0xFF) ^ ((comment_crc >> 8) & 0xFF) ^ ((comment_crc >> 16) & 0xFF) ^ ((comment_crc >> 24) & 0xFF));
    if (comment_crc_xored == 0) comment_crc_xored++;
    if ( comment_crc_xored != *(pdata.comment_field+comment_field_length-1) ) {         /* Comment field CRC is not OK */
        free_data(&pdata);
        inflate_st(filename);
        return;
    }
    if (opt_perf_stat) stat_start_clocks = ippGetCpuClocks();
    /* Comment field is OK for multi-threading inflate */
#if _MSC_VER >= 1400  && !defined(_AMD64_)
    sscanf_s((const char*)pdata.comment_field+1, "%1X", &chunks);
#else
    sscanf((const char*)pdata.comment_field+1, "%1X", &chunks);
#endif
    if ( (chunks-1)*16+3 != comment_field_length ) {          /* Suspicious comment field */
        free_data(&pdata);
        inflate_st(filename);
        return;
    }
    /* Restore chunk offset and chunk CRC arrays */
    pdata.offset_array = (long *) zalloc(sizeof(long)*(chunks-1));
    pdata.crc_array = (Ipp32u *) zalloc(sizeof(long)*(chunks-1));
    ptr = pdata.comment_field+2;
#if _MSC_VER >= 1400  && !defined(_AMD64_)
    for ( i=0; i<chunks-1; i++ ) ptr += (sscanf_s((const char*)ptr, OFFSET_FORMAT, pdata.offset_array+i)*8);
    for ( i=0; i<chunks-1; i++ ) ptr += (sscanf_s((const char*)ptr, OFFSET_FORMAT, pdata.crc_array+i)*8);
#else
    for ( i=0; i<chunks-1; i++ ) ptr += (sscanf((const char*)ptr, OFFSET_FORMAT, pdata.offset_array+i)*8);
    for ( i=0; i<chunks-1; i++ ) ptr += (sscanf((const char*)ptr, OFFSET_FORMAT, pdata.crc_array+i)*8);
#endif
    if (!opt_test) {
    /* Create output file */
    if (!opt_name || (flag_byte & FNAME)==0) {       /* Use don't want to restore original name */
        pdata.result_fname = (char*)zalloc(strlen((const char*)filename)+1);
        if (!get_nosuf_name(filename, pdata.result_fname)) {
            free_data(&pdata);
            return;
    }
        } else
        restore_output_name(filename, pdata.orig_fname, pdata.result_fname);
        if (!opt_to_stdout) {
            if ( (fd_out = out_open(pdata.result_fname, zsizeuncompressed(pdata.fd_in))) == NULL ) {
                fprintf(stderr, "Cannot open output file\n");
                error_file(pdata.result_fname);
                free_data(&pdata);
                inflate_st(filename);
                return;
            }
            fd_out->uncompressed_size = targetFileSize = zsizeuncompressed(pdata.fd_in);
        } else
            fd_out = zopen_stdout();
        if (opt_io_write_method == METHOD_LOW)
            tmp_fd_vector = (GZIP_FILE**)zalloc(sizeof(GZIP_FILE*) * chunks);   /* For low-level writes we need temporary files */
    }
    if (opt_test && opt_verbose) fprintf(stderr, "%s:\t ", filename);
    error_found = 0;
#ifdef GZIP_PTHREADS
    /* Here we need to use correct number of threads. Number of threads created must not be greater than number of processors available */
    p_params = (struct ic_params*)zalloc(chunks * sizeof(struct ic_params));
    pthreads = (pthread_t*)zalloc(num_threads * sizeof(pthread_t));
    for (i=0; i<num_threads; i++) {
        pthreads[i] = (pthread_t)NULL;
    }
    pthread_attr_init(&attr);
    pthread_cond_init(&condition_cond, NULL);
    for (i=0; i<chunks; i++) {
        if ( i == 0 ) {                   /* If initial chunk */
            p_params[i].from_crc = 0L;
            p_params[i].to_crc = pdata.crc_array[0];
            p_params[i].from_pos = inptr-pdata.inbuf;
            p_params[i].to_pos = pdata.offset_array[0];
        } else if ( i == chunks-1 ) {     /* If last chunk */
            p_params[i].from_crc = pdata.crc_array[i-1];
            p_params[i].to_crc = 0L;
            p_params[i].from_pos = pdata.offset_array[i-1];
            p_params[i].to_pos = file_info.st_size; /* Don't read last 8 bytes (4 bytes of CRC and 4 bytes of encoded file length */
        } else {
            p_params[i].from_crc = pdata.crc_array[i-1];
            p_params[i].to_crc = pdata.crc_array[i];
            p_params[i].from_pos = pdata.offset_array[i-1];
            p_params[i].to_pos = pdata.offset_array[i];
        }
        if (opt_io_write_method == METHOD_LOW && !opt_test) {
            if (i > 0) {
                char tmp_filename[MAX_NAME];
                if ( (tmp_fd_vector[i] = ZOPEN(mk_chunk_name(filename, i, tmp_filename), GZIP_FILE_WRITE, zsizeuncompressed(pdata.fd_in))) == NULL ) {
                    perror("cannot create temporary file");
                        free_data(&pdata);
                        return;
                }
            } else {
                tmp_fd_vector[i] = fd_out;
            }
            p_params[i].tmp_fd = tmp_fd_vector[i];
        }
        p_params[i].in_buffer_size = insize;
        p_params[i].out_buffer_size = outsize;
        p_params[i].filename = filename;
        p_params[i].chunk0_inptr = inptr;
        p_params[i].chunk0_inbuf = pdata.inbuf;
        p_params[i].chunk0_fd_in = pdata.fd_in;
        p_params[i].err_status = &error_found;
        p_params[i].chunk = i;
        p_params[i].max_chunk = chunks;
        p_params[i].targetFileSize = targetFileSize;
        if (opt_io_write_method == METHOD_MMAP && !opt_test) p_params[i].tmp_fd = fd_out;  /* In MMAP we write to the same output file in parallel */
        if (next_free_thread_slot < num_threads) {
            pthread_create(&pthreads[next_free_thread_slot++], &attr, inflate_chunk, &p_params[i]);
        } else {
            int j;
            pthread_mutex_lock(&condition_mutex);
            pthread_cond_wait(&condition_cond, &condition_mutex);   /* Wait for any pthread to finish */
            pthread_join(f_pthread, NULL);                          /* f_pthread must contain identifier of finishing pthread */
            for (j=0; j<num_threads; j++) {
                if (pthread_equal(f_pthread, pthreads[j])) {         /* This thread finished. start new pthread */
                    pthread_create(&pthreads[j], &attr, inflate_chunk, &p_params[i]);
                    break;
                }
            }
            pthread_mutex_unlock(&condition_mutex);                 /* Free other pthreads to finish */
        }
    }
    for (i=0; i<num_threads; i++) {
        if (pthreads[i] != (pthread_t)NULL) {
            pthread_join(pthreads[i], NULL);
        }
    }
#else       /* GZIP_PTHREADS */
#pragma omp parallel num_threads(chunks) private(from_crc, to_crc, from_pos, to_pos)
    {
        int id = omp_get_thread_num();      /* Get current thread order number */

        if ( id == 0 ) {                   /* If initial chunk */
            from_crc = 0L;
            to_crc = pdata.crc_array[0];
            from_pos = (unsigned long)(inptr-pdata.inbuf);
            to_pos = pdata.offset_array[0];
        } else if ( id == chunks-1 ) {     /* If last chunk */
            from_crc = pdata.crc_array[id-1];
            to_crc = 0L;
            from_pos = pdata.offset_array[id-1];
            to_pos = file_info.st_size; /* Don't read last 8 bytes (4 bytes of CRC and 4 bytes of encoded file length */
        } else {
            from_crc = pdata.crc_array[id-1];
            to_crc = pdata.crc_array[id];
            from_pos = pdata.offset_array[id-1];
            to_pos = pdata.offset_array[id];
        }
        if (opt_io_write_method == METHOD_LOW && !opt_test) {
        /* We need to create temporary output files to store intermediate data */
            if (id > 0) {
                char tmp_filename[MAX_NAME];
                    if ( (tmp_fd_vector[id] = ZOPEN(mk_chunk_name(filename, id, tmp_filename), GZIP_FILE_WRITE, zsizeuncompressed(pdata.fd_in))) 
                         == NULL ) {
                    perror("cannot create temporary file");
                        free_data(&pdata);
                    abort();
                }
            } else 
                tmp_fd_vector[id] = fd_out;
            }
        if (opt_io_write_method == METHOD_MMAP) 
            inflate_chunk(id, from_crc, to_crc, from_pos, to_pos, insize, outsize, filename, inptr, 
                          pdata.inbuf, pdata.fd_in, fd_out, &error_found, targetFileSize);
        else
            inflate_chunk(id, from_crc, to_crc, from_pos, to_pos, insize, outsize, filename, inptr, 
                          pdata.inbuf, pdata.fd_in, tmp_fd_vector[id], &error_found, targetFileSize);
    }   /* End of omp parallel section */
#endif              /* GZIP_PTHREADS */
    if (opt_io_write_method == METHOD_LOW) {
        /* Allocate lengthy buffer for tmp file re-read */
        zfree(pdata.inbuf);
        pdata.inbuf = (UBYTE*)zalloc(INBUF_SIZE);
        insize = INBUF_SIZE;
    }
    if (!opt_test) {
        if (opt_io_write_method == METHOD_LOW) {
            for (i=1; i<chunks; i++) {
                if ( zseek(tmp_fd_vector[i], 0, 0)) {
                    perror("cannot rewind temporary file"); abort();
                }
                while ( !zeof(tmp_fd_vector[i]) ) {
                    incnt = zread(tmp_fd_vector[i], pdata.inbuf, insize);
                    zwrite(fd_out, pdata.inbuf, incnt);
                }
                zclose(tmp_fd_vector[i]);
                remove_tmp_file(filename, i);
            }
        } /* opt_io_write_method == METHOD_LOW*/
    } /* !opt_test */
    /* Free all */
    /* Print out decompression statistics */
    if (opt_perf_stat && !opt_test && !opt_to_stdout && filename != NULL) {
        stat_finish_clocks = ippGetCpuClocks();
        fprintf(stderr, "%s:\t inflating ", filename);
        print_io_methods();
        fprintf(stderr, " MT -- %6.1f clocks per input symbol\n", (float)(stat_finish_clocks - stat_start_clocks) / pdata.fd_in->size);
    }
    if (opt_verbose && !opt_test && !opt_to_stdout && filename != NULL) {
        fprintf(stderr, "%s:\t %6.1f%% -- replaced with %s\n", filename, 100*(1.-(float)pdata.fd_in->size/fd_out->bytes_out), pdata.result_fname);
    }
    if (!opt_test) {
        zclose(fd_out);
    }
    if (opt_io_write_method == METHOD_LOW) {
    zfree(tmp_fd_vector);
    }
    if (!opt_notime && !opt_test)          /* Need to restore original file creation time */
        zcopytime(filename, pdata.result_fname, timestamp);
    if (opt_test && opt_verbose && !error_found) {
        fprintf(stderr, "OK\n");
    }
    free_data(&pdata);
#ifdef GZIP_PTHREADS
    zfree(p_params);
    zfree(pthreads);
    pthread_attr_destroy(&attr);
#endif
    if (!opt_to_stdout && filename != NULL && !opt_test) {
        if (UNLINK(filename) != 0) {
            perror("cannot unlink");
        }
    }
}
/**
 * uncompress chunk in parallel thread
 * @param chunk - current chunk number
 * @param from_crc - initial CRC32 value
 * @param to_crc - final CRC32 value
 * @param from_pos - initial file position of compressed chunk
 * @param to_pos - final file position
 * @param in_buffer_size - input buffer size
 * @param out_buffer_size - output buffer size
 * @param filename - input file name
 * @param chunk0_inptr - current ptr in outer buffer to be used
 *                     in chunk #0
 * @param chunk0_inbuf - input outer buffer to be used in chunk
 *                     #0
 * @param chunk0_fd_in - input file descriptor to be used in
 *                     chunk #0
 * 
 * @return int - 1 (success) or 0 (failure)
 */

static char invalid_crc_format_string[] = "\n%s: %s: invalid compressed data--crc error, chunk %d\n";

#ifdef GZIP_PTHREADS
void *inflate_chunk(void *params)
#else
int inflate_chunk(int chunk, Ipp32u from_crc, Ipp32u to_crc, unsigned long from_pos, unsigned long to_pos, 
                           size_t in_buffer_size, size_t out_buffer_size, char *filename, 
                           UBYTE *chunk0_inptr, UBYTE *chunk0_inbuf, GZIP_FILE *chunk0_fd_in,
                           GZIP_FILE *tmp_fd, int *err_status, int targetFileSize)
#endif
{
    UBYTE *inbuf;
    UBYTE *inptr;
    size_t incnt;                  /* No. of actual bytes read */
    UBYTE *outbuf;
    UBYTE *outptr;
    size_t outcnt;
    GZIP_FILE *fd_in;
    size_t read_size;
    size_t bytes_left;             /* How much to read yet */
    IppLZ77State_8u *ippState;
    IppStatus status;
    IppLZ77HuffMode huff_mode;
    IppLZ77Flush flush_mode;
    Ipp32u crc, stored_crc;
#ifdef GZIP_PTHREADS
    struct ic_params *pic;
    int chunk;
    Ipp32u from_crc;
    Ipp32u to_crc;
    long from_pos;
    long to_pos;
    size_t in_buffer_size;
    size_t out_buffer_size;
    char *filename;
    UBYTE *chunk0_inptr;
    UBYTE *chunk0_inbuf;
    GZIP_FILE *chunk0_fd_in;
    GZIP_FILE *tmp_fd;
    int *err_status;
    int targetFileSize;
    int max_chunks;
#endif
    DWORD targetFilePosition;
    GZIP_VIEW *pView;

#ifdef GZIP_PTHREADS
    pic = (struct ic_params*)params;
    chunk = pic->chunk;
    max_chunks = pic->max_chunk;
    from_crc = pic->from_crc;
    to_crc = pic->to_crc;
    from_pos = pic->from_pos;
    to_pos = pic->to_pos;
    in_buffer_size = pic->in_buffer_size;
    out_buffer_size = pic->out_buffer_size;
    filename = pic->filename;
    chunk0_inptr = pic->chunk0_inptr;
    chunk0_inbuf = pic->chunk0_inbuf;
    chunk0_fd_in = pic->chunk0_fd_in;
    tmp_fd = pic->tmp_fd;
    err_status = pic->err_status;
    targetFileSize = pic->targetFileSize;
#endif
    if ( chunk == 0 ) {           /* initial chunk */
        inbuf = chunk0_inbuf;
        inptr = chunk0_inptr;
        fd_in = chunk0_fd_in;
        if (opt_io_write_method == METHOD_MMAP && !opt_test) {
            if ((pView = zopenview(tmp_fd, 0)) == NULL) {
                error_system("Cannot open view for address 0");
                return 0;
            }
        }
    } else {
        /* Open input file in parallel thread */
        if ( (fd_in = ZOPEN(filename, GZIP_FILE_READ, 0)) == NULL ) {
            fprintf(stderr, "Cannot open input file in inflate_chunk\n");
            error_file(filename);
            return 0;
        }
        /* Allocate input buffer */
        inbuf = (UBYTE *) zalloc(in_buffer_size);
        inptr = inbuf;
        /* This is not chunk#0, so we need to seek to desired position and to read input data */
        if ( zseek(fd_in, from_pos, SEEK_SET) ) {
            fprintf(stderr, "Cannot seek in inflate_chunk\n");
            error_file(filename);
            if ( chunk ) {
                zfree(inbuf);
                zclose(fd_in);
            }
            return 0;
        }
        /* For writing to the same file in parallel mode we need to set file position for output and lock target region of output file */
#ifndef GZIP_PTHREADS
        targetFilePosition = (targetFileSize / omp_get_num_threads()) * chunk;
#else
        targetFilePosition = (targetFileSize * chunk) / max_chunks;
#endif
        if (opt_io_write_method == METHOD_MMAP && !opt_test) {
            if ((pView = zopenview(tmp_fd, targetFilePosition)) == NULL) {
                zfree(inbuf);
                zclose(fd_in);
                return 0;
            }
        }
    }
    /* Create output buffer */
    outbuf = (UBYTE *)zalloc(out_buffer_size);
    outptr = outbuf;
    outcnt = out_buffer_size;
    bytes_left = to_pos - from_pos;
    /* Create IPP DC structure */
    status = ippsDecodeLZ77InitAlloc_8u(IppLZ77CRC32, &ippState);
    status = ippsDecodeLZ77SetStatus_8u(IppLZ77InflateStatusInit, ippState);
    status = ippsDecodeLZ77SetStatus_8u(IppLZ77InflateStatusHuffProcess, ippState);
    /* Determine how much to read */
    read_size = MINSIZE(in_buffer_size, to_pos-from_pos);
    if ( chunk ) {       /* If not initial chunk - read */
      incnt = zread(fd_in, inbuf, read_size);
    } else
        incnt = (int)MINSIZE(to_pos-from_pos, in_buffer_size-from_pos);
    crc = from_crc;
    flush_mode = IppLZ77NoFlush;
    bytes_left -= incnt;
    do {
        status = ippsDecodeLZ77GetBlockType_8u(&inptr, (int*)&incnt, &huff_mode, ippState);
        if ( status == ippStsNoErr ) {
            do {
                switch (huff_mode) {
                case IppLZ77UseFixed:
                    status = ippsDecodeLZ77FixedHuffFull_8u(&inptr, (int*)&incnt, &outptr, (int*)&outcnt, flush_mode, ippState);
                    break;
                case IppLZ77UseDynamic:
                    status = ippsDecodeLZ77DynamicHuffFull_8u(&inptr, (int*)&incnt, &outptr, (int*)&outcnt, flush_mode, ippState);
                    break;
                case IppLZ77UseStored:
                    status = ippsDecodeLZ77StoredBlock_8u(&inptr, (int*)&incnt, &outptr, (int*)&outcnt, ippState);
                    break;
                default:
                    fprintf(stderr, "%s: %s - unknown Huffman mode %d\n", prog_name, (char*)filename, (int)huff_mode);
                    break;
                }
                switch (status) {
                case ippStsSrcSizeLessExpected:
                case ippStsNoErr:
                    if (bytes_left > 0 && incnt == 0) {
                        read_size = MINSIZE(in_buffer_size, bytes_left);
                        incnt = zread(fd_in, inbuf, read_size);
                        inptr = inbuf;
                        bytes_left -= incnt;
                    }
                    break;
                case ippStsDstSizeLessExpected:
                    ippsCRC32_8u(outbuf, (int)(out_buffer_size-outcnt), &crc);
                    if (!opt_test) {
                        if (opt_io_write_method == METHOD_MMAP) {
                            if (zwriteview(pView, outbuf, out_buffer_size-outcnt) != (out_buffer_size-outcnt) ) {
                                perror("error writing to temporary file");
                                zfree(outbuf);
                                if ( chunk ) {
                                    zfree(inbuf);
                                    zclose(fd_in);
                                }
                                zcloseview(pView);
                                return 0;
                            }
                        } else {    /* METHOD == LOW */
                        if (zwrite(tmp_fd, outbuf, out_buffer_size-outcnt) != (out_buffer_size-outcnt) ) {
                            perror("error writing to temporary file");
                            zfree(outbuf);
                            if ( chunk ) {
                                zfree(inbuf);
                                zclose(fd_in);
                            }
                            zclose(tmp_fd);
                            return 0;
                        }
                    }
                    }
                    outptr = outbuf;
                    outcnt = out_buffer_size;
                    break;
                case ippStsStreamEnd:
                    goto out_of_cycle;
                    break;
                default:
                    fprintf(stderr, "%s: %s - unexpected IPP DC status - ", prog_name, (char*)filename);
                    print_ipp_status(status);
#ifdef GZIP_PTHREADS
                    return NULL;
#endif
                    abort();
                }
            } while ( status != ippStsNoErr );  /* Do until LZ77 block is done */
        } /* GetBlockType */
        if (status != ippStsSrcSizeLessExpected && status != ippStsNoErr) {
            print_ipp_status(status);
            break;
        }
        if (bytes_left > 0 && incnt == 0) {
            /* Read next portion of input */
            read_size = MINSIZE(in_buffer_size, bytes_left);
            incnt = zread(fd_in, inbuf, read_size);
            inptr = inbuf;
            bytes_left -= incnt;
        }
    } while (bytes_left > 0 || incnt > 0);
out_of_cycle:
    if ((out_buffer_size-outcnt) != 0) {
        ippsCRC32_8u(outbuf, (int)(out_buffer_size-outcnt), &crc);
        if (!opt_test) {
            if (opt_io_write_method == METHOD_MMAP) {
                if ( zwriteview(pView, outbuf, out_buffer_size-outcnt) != (out_buffer_size-outcnt) ) {
                    perror("error writing to temporary file");
                    zfree(outbuf);
                    if ( chunk ) {
                        zfree(inbuf);
                        zclose(fd_in);
                    }
                    zcloseview(pView);
                    return 0;
                }
            } else {    /* METHOD == LOW */
            if ( zwrite(tmp_fd, outbuf, out_buffer_size-outcnt) != (out_buffer_size-outcnt) ) {
                perror("error writing to temporary file");
                zfree(outbuf);
                if ( chunk ) {
                    zfree(inbuf);
                    zclose(fd_in);
                }
                zclose(tmp_fd);
                return 0;
            }
        }
        }
        outptr = outbuf;
        outcnt = out_buffer_size;
    }
    if (opt_io_write_method == METHOD_MMAP && !opt_test) 
        zcloseview(pView);
    if (to_crc != 0) {
        if ( crc != to_crc ) {        /* CRC error for current chunk */
            fprintf(stderr, invalid_crc_format_string, prog_name, filename, chunk);
            *err_status = 1;
            if (chunk) {
                zfree(inbuf);
                zclose(fd_in);
            }
            zfree(outbuf);
            return 0;
        } else {
        }
    } else {
        stored_crc = GET_LONG(inptr); inptr += 4;
        if (crc != stored_crc) {
            fprintf(stderr, invalid_crc_format_string, prog_name, filename, chunk);
            *err_status = 1;
            if (chunk) {
                zfree(inbuf);
                zclose(fd_in);
            }
            zfree(outbuf);
            return 0;
        }
    }
/* End of input buffer processing */
    ippsLZ77Free_8u(ippState);
    if (chunk) {
        zfree(inbuf);
        zclose(fd_in);
    }
    zfree(outbuf);
#ifdef GZIP_PTHREADS
    pthread_mutex_lock(&condition_mutex);       /* Block other threads from going further */
    f_pthread = pthread_self();                 /* Get self thread identifier */
    pthread_cond_signal(&condition_cond);       /* Signal to master thread to proceed */
    pthread_mutex_unlock(&condition_mutex);     /* Allow master thread to proceed */
    return NULL;
#else
    return 1;
#endif
}

#endif /* Parallel condition compilation */
