/*
//
//                  INTEL CORPORATION PROPRIETARY INFORMATION
//     This software is supplied under the terms of a license agreement or
//     nondisclosure agreement with Intel Corporation and may not be copied
//     or disclosed except in accordance with the terms of that agreement.
//          Copyright(c) 2007-2008 Intel Corporation. All Rights Reserved.
//
//     Intel(R) Integrated Performance Primitives Data Compression Domain-based GZIP
//
//  By downloading and installing this sample, you hereby agree that the
//  accompanying Materials are being provided to you under the terms and
//  conditions of the End User License Agreement for the Intel(R) Integrated
//  Performance Primitives product previously accepted by you. Please refer
//  to the file ippEULA.rtf located in the root directory of your Intel(R) IPP
//  product installation for more information.
//
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <ctype.h>
#ifndef _WIN32
    #include <unistd.h>
    #ifdef __MACH__
        #include <sys/types.h>
        #include <sys/sysctl.h>
    #endif
#endif
#ifndef GZIP_PTHREADS
    #if defined(_OPENMP) && defined(GZIP_PARALLEL)
    #include <omp.h>
#endif
#endif

#include <ippcore.h>

#include "ipp_gzip.h"

/*! \brief In Windows possible filename separator can be \ */
#define FNAME_SEPARATOR_1 '\\'
/*! \brief or / */
#define FNAME_SEPARATOR_2 '/'
/*! \brief Defines max pointer of two */
#define maxptr(a, b) ((a > b)? a : b)

static char *gzip_suffixes[] = {
    ".gz", ".taz", ".tgz", ".z"
};

/**
 * Prints current program name on stderr
 */
void print_progname()
{
    fprintf(stderr, "%s: ", prog_name);
}
/**
 * Produces "progname: text" message on stderr
 * @param text
 */
void warn_message(char *text)
{
    print_progname();
    fprintf(stderr, "%s\n", text);
}
/**
 * Program name separation function
 * @param full_name
 * 
 * @return char*
 */
char *get_program_name(char *full_name)
{
    char *ptr1, *ptr2;

    ptr1 = strrchr(full_name, FNAME_SEPARATOR_1);
    ptr2 = strrchr(full_name, FNAME_SEPARATOR_2);
    if (!ptr1 && !ptr2) 
        return full_name;
    return maxptr(ptr1, ptr2)+1;
}
/**
 * Error during file i/o occurred
 * @param filename
 */
void error_file(char *filename)
{
    print_progname();
    perror((const char*)filename);
}
/**
 * Wrapper around standard malloc to shout about memory lack
 * @param size
 * 
 * @return void*
 */
void *zalloc(size_t size)
{
    void *ptr;

    ptr = malloc(size);
    if (ptr == NULL) {
        warn_message("cannot allocate memory");
        exit(-1);
    }
    return ptr;
}
/**
 * Counterpart for zalloc function
 * @param ptr
 */
void zfree(void *ptr)
{
    if (ptr == NULL)
        return;
    free(ptr);
}
/**
 * Constructs output filename from input filename and possible
 * new suffix if specified by command line option
 * @param i_name - input file name
 * @param ou_name - output file name
 */
void create_output_name(char *i_name, char *ou_name)
{
    strcpy(ou_name, i_name);
    if (!opt_new_suffix)                /* If new suffix is not specified */
        strcat(ou_name, ".gz");
    else
        strcat(ou_name, suffix);        /* Add new suffix specified in command line */
}
/**
 * Creates standard gzip-like header and returns pointer to end
 * of header (before possible comment field)
 * @param out - pointer to output buffer
 * @param filename - input filename
 * @param istat - file statistics structure
 * @param comment_needed - 1 if we need to put comment into
 *                       header
 * 
 * @return char*
 */
#if _MSC_VER >= 1400  && !defined(_AMD64_)
UBYTE *create_gzip_hdr(UBYTE *out, char *filename, struct _stat *istat, int comment_needed)
#else
UBYTE *create_gzip_hdr(UBYTE *out, char *filename, struct stat *istat, int comment_needed)
#endif
{
    UBYTE flag_byte, xfl_byte;

    *out++ = GZIP_HDR_1;
    *out++ = GZIP_HDR_2;
    *out++ = DEFLATE;
    flag_byte = 0;
    if (!opt_noname) {
        flag_byte |= FNAME;
    }
    if (comment_needed) {
        flag_byte |= FCOMMENT;
    }
    *out++ = flag_byte;
    PUT_LONG(out, istat->st_mtime);

    xfl_byte = 0;       /* Extra flags */
    if (opt_compress_fast) {
        xfl_byte |= XFL_FAST;
    }
    if (opt_compress_best) {
        xfl_byte |= XFL_BEST;
    }
    *out++ = xfl_byte;
    *out++ = OS_BYTE;
    if (!opt_noname) {              /* We must put original filename into the header */
        while (*out++ = *filename++)
            ;
    }
    return out;
}

/*! \brief Contains options description */
struct option_descr {       
 char *long_name;               /**< Long option name */
 char short_name;               /**< Option character */
 short *flag_to_set;     /**< Integer flag to set if option encountered */
 char **modifier;               /**< Additional string to option */
 char *description;             /**< Explanation for help() function */
};
/*! \brief Array containing all available option descriptions */
struct option_descr all_options[] = {
/*    { "ascii",      'a', &opt_ascii, NULL,           "make text mode ASCII (NOT IMPLEMENTED)" }, */
    { "to-stdout",  'c', &opt_to_stdout, NULL,   "output to stdout" },
    { "stdout",     'c', &opt_to_stdout, NULL,      "       the same" },
    { "decompress", 'd', &opt_decompress, NULL, "decompress" },
    { "uncompress", 'd', &opt_decompress, NULL, "       the same" },
/*    { "encrypt",    'e', &opt_encrypt, NULL,       "encrypt (NOT IMPLEMENTED)" }, */
    { "force",      'f', &opt_force, NULL,           "force output file overwrite" },
    { "help",       'h', &opt_help, NULL,             "print this text" },
/*    { "pkzip",      'k', &opt_pkzip, NULL,           "force PKZIP compatibility (NOT IMPLEMENTED)" }, */
    { "list",       'l', &opt_list, NULL,             "list content of gzip archive" },
/*    { "license",    'L', &opt_license, NULL,       "print license message" }, */
    { "no-name",    'n', &opt_noname, NULL,        "don't save/restore original filename/time" },
    { "name",       'N', &opt_name, NULL,             "save/restore original filename/time" },
/*    { "quiet",      'q', &opt_quiet, NULL,           "quiet/silent mode" }, */
/*    { "silent",     'q', &opt_quiet, NULL,          "       the same" }, */
    { "recursive",  'r', &opt_reqursive, NULL,   "recurse into directories" },
    { "suffix",     'S', &opt_new_suffix, &suffix,     "make new suffix instead of .gz" },
    { "test",       't', &opt_test, NULL,             "test archive integrity" },
    { "no-time",    'T', &opt_notime, NULL,        "don't save/restore file date/timestamp" },
    { "verbose",    'v', &opt_verbose, NULL,       "print some intermediate information" },
    { "version",    'V', &opt_version, NULL,       "print current version number" },
    { "fast",       '1', &opt_compress_fast, NULL,    "use faster compression method" },
    { "best",       '9', &opt_compress_best, NULL,    "use best compression method" },
/*    { "lzw",        'Z', &opt_lzw_compat, NULL,        "enable LZW compatibility (NOT IMPLEMENTED" }, */
/*    { "bits",       'b', &opt_bits, NULL,             "make number of bits per source byte" }, */
/*    { "rsyncable",  'R', &opt_rsyncable, NULL,   "make rsync-enabled archive" } */

    /* Debugging options */
    { "num-threads", 'm', &opt_dbg_multithreads, &opt_num_threads_forced, "DEBUG: set number of threads to create"},
    { "force-dynamic", 'b', &opt_force_dynamic, NULL, "force dynamic Huffman coding"},
    { "min-size",       'j',    &opt_dbg_slice_size,    &opt_slice_size_array,      
                            "DEBUG: set minimum input file length to slice\nInput/Ouput options:"},
    { "DEBUG",          'D', &opt_debug, NULL, "DEBUG mode: prints some debug info"},
    /* Input/output options */
    { "read-method",    'y',    &opt_io_read_method,    &ptr_read_method,   "I/O: read method to use -y 0 (low/level) -y 1 (mmap)"},
    { "write-method",   'w',    &opt_io_write_method,   &ptr_write_method,  "I/O: write method to use -w 0 (low/level) -w 1 (mmap)"},
    { "read-buffer",    'u',    &opt_io_read_buffer,    &ptr_read_buffer,   "I/O: read buffer size in Kbytes"},
    { "write-buffer",   'i',    &opt_io_write_buffer,   &ptr_write_buffer,  "I/O: write buffer size in Kbytes"},
    { "stat",           's',    &opt_perf_stat,         NULL,               "display performance statistics"}
};
/**
 * Prints list of options currently available
 */
void help()
{
    int i;

    fprintf(stderr, "Usage: %s [OPTION]... [FILE]...\nCompress or uncompress FILEs (by default, compress FILES in-place)\n\n", prog_name);
    fprintf(stderr, "Mandatory arguments to long options are mandatory for short options too.\n");
    for (i=0; i<sizeof(all_options)/sizeof(struct option_descr); i++) {
        fprintf(stderr, "\t-%c, --%s\t%s\n", all_options[i].short_name, all_options[i].long_name, all_options[i].description);
    }
}

static char unknown_option[] = "Unknown option '%s', use '-h' for help\n";  /**< Unknown option message format string */

/**
 * Processes all command line options, returns index of first
 * filename in command line filename argument
 * @param argc
 * @param argv
 * 
 * @return int
 */
int get_options(int argc, char *argv[])
{
    int i, j;
    char *ptr;
    unsigned short option_ok;

    for (i=1; i<argc; i++) {
        ptr = argv[i];
        if (*ptr == '-' && *(ptr+1) == '-') {           /* long option name found */
            option_ok = 0;
            for (j=0; j<sizeof(all_options)/sizeof(struct option_descr); j++) {
                if (strcmp(all_options[j].long_name, ptr+2) == 0) {
                    (*all_options[j].flag_to_set)++;
                    if (all_options[j].modifier) {
                        *all_options[j].modifier = argv[++i];
                    }
                    option_ok = 1;
                    break;
                }
            }
            if (!option_ok) {
                fprintf(stderr, unknown_option, argv[i]);
            }
        } else if (*ptr == '-') {                       /* short option name */
            for (j=0; j<sizeof(all_options)/sizeof(struct option_descr); j++) {
                option_ok = 0;
                if (*(ptr+1) == all_options[j].short_name) {
                    (*all_options[j].flag_to_set)++;
                    if (all_options[j].modifier) {
                        *all_options[j].modifier = argv[++i];
                    }
                    option_ok = 1;
                    break;
                }
            }
            if (!option_ok) {
                fprintf(stderr, unknown_option, argv[i]);
            }
        } else {                                    /* start of filenames */
            return i;
        }
    }
    return argc;        /* no filename found */
}
/**
 * Get basename (no paths, no .gz extension) from filename.
 * Store basename in "basename"
 * @param filename
 * @param basename
 */
void get_base_name(char *filename, char *basename)
{
    char *name_start;
    char *ptr;

    ptr = strrchr((const char*)filename, FNAME_SEPARATOR_1);
    name_start = maxptr(ptr, strrchr((const char*)filename, FNAME_SEPARATOR_2));
    if (name_start == NULL) {   /* i.e. no path separators in the name */
        name_start = filename;
    } else {
        name_start++;
    }
    *basename = '\0';
    for (ptr=name_start; *ptr; ptr++) {
        if (strcmp((const char*)ptr, ".gz") == 0) {
            *basename = '\0';
            return;
        } else if (strcmp((const char*)ptr, ".tgz") == 0 || strcmp((const char*)ptr, ".taz")==0) {
            strcpy(basename, ".tar");
            return;
        }
        *basename++ = *ptr;
    }
    *basename = '\0';
    return;
}
/**
 * Returns filename without suffix if suffix is known
 * (.gz.tgz.etc). Returns 0 if suffix is not known.
 * @param filename
 * @param nosuf_name
 * 
 * @return int
 */
int get_nosuf_name(char *filename, char *nosuf_name)
{
    char *dot;
    int i;

    strcpy(nosuf_name, filename);
    for (i=0; i<sizeof(gzip_suffixes)/sizeof(char*); i++) {
        size_t str_size;
        dot = strstr(nosuf_name, gzip_suffixes[i]);
        str_size = strlen(dot);
        if (dot != NULL &&  str_size == strlen(gzip_suffixes[i])) {
            *dot = '\0';        /* Mark the end */
    return 1;
}
    }
    return 0;
}
/**
 * Restore basename for output file
 * Take input filename and substitute it with filename stored in
 * gzip archive
 * @param filename
 * @param orig_name
 * @param mod_name
 */
void restore_output_name(char *filename, char *orig_name, char *mod_name)
{
    char *fname_separator;
    char *ptr;

    strcpy(mod_name, filename);
    ptr = strrchr((const char*)mod_name, FNAME_SEPARATOR_1);
    fname_separator = maxptr(ptr, strrchr((const char*)mod_name, FNAME_SEPARATOR_2));
    if (fname_separator == NULL) {      /* No directory name is specified */
        strcpy(mod_name, orig_name);
    } else {
        fname_separator++;
        *fname_separator = '\0';        /* Cut filename */
        strcat(mod_name, orig_name);
    }
}
/**
 * 
 * @param template_name
 * @param chunk
 * @param output_name
 * 
 * @return char*
 */
char *mk_chunk_name(char *template_name, int chunk, char *output_name)
{
#ifndef _WIN32
    char basename[PATH_MAX];
    get_base_name(template_name, basename);
    sprintf(output_name, "/tmp/%s.%02d", basename, chunk);
#else
    sprintf(output_name, "%s.%02d", template_name, chunk);
#endif
    return output_name;
}
int remove_tmp_file(char *template_name, int chunk)
{
    int status;
    char tmp_name[MAX_NAME];

    status = UNLINK(mk_chunk_name(template_name, chunk, tmp_name));
    if (status == -1)
        perror("Remove file");
    return status;
}
/**
 * Print status of last call to IPP library
 * @param status
 */
void print_ipp_status(IppStatus status)
{
    fprintf(stderr, "IPP status: %s\n", ippGetStatusString(status));
}
/**
 * Create output file.
 * @param filename
 * 
 * @return FILE* - file stream of open file. -1 if not open.
 */
GZIP_FILE *out_open(char *filename, size_t initial_size)
{
    int answer;
#if _MSC_VER >= 1400  && !defined(_AMD64_)
    struct _stat chk_stat;
#else
    struct stat chk_stat;
#endif    

    if (opt_force)
        return ZOPEN(filename, GZIP_FILE_WRITE, initial_size);
#if _MSC_VER >= 1400  && !defined(_AMD64_)
    if (_stat(filename, &chk_stat) == 0) {
#else
    if (stat(filename, &chk_stat) == 0) {
#endif
        fprintf(stderr, "%s: %s already exists; do you wish to overwrite (y or n)? ", prog_name, (char*)filename);
        answer = fgetc(stdin);
        if (tolower(answer) != 'y') {
            fprintf(stderr, "        not overwritten\n");
            return NULL;
        }
    }
    return ZOPEN(filename, GZIP_FILE_WRITE, initial_size);
}
static char *list_title = "compressed        uncompressed  ratio uncompressed_name";
static char *list_title_verbose = "method  crc     date  time           compressed        uncompressed  ratio uncompressed_name";
static char *year="JanFebMarAprMayJunJulAugSepOctNovDec";
static short list_header = 0;
void show_list(char *filename , GZIP_FILE *file, long time_stamp)
{
    Ipp32u file_crc;
    long file_len;
    UBYTE file_tail[8];
    UBYTE *tail_ptr;
#if _MSC_VER >= 1400  && !defined(_AMD64_)
    struct _stat file_info;
#else
    struct stat file_info;
#endif
    char orig_fname[MAX_NAME];
#if _MSC_VER >= 1400  && !defined(_AMD64_)
    struct tm ltime;
#else
    time_t cvtime;
    struct tm *ltime;
#endif

#if _MSC_VER >= 1400  && !defined(_AMD64_)
    if (_stat((const char*)filename, &file_info) == (-1L)) {
#else
    if (stat((const char*)filename, &file_info) == (-1L)) {
#endif
        return;
    }
    if (zseek(file, -8, SEEK_END) == 0) {
        error_file(filename); 
        return;
    }
    if (zread(file, file_tail, 8) != 8) {
        error_file(filename);
        return;
    }
    tail_ptr = file_tail;
    file_crc = GET_LONG(tail_ptr); tail_ptr += 4;
    file_len = GET_LONG(tail_ptr); tail_ptr += 4;
    get_base_name(filename, orig_fname);
    if (opt_verbose) {
#if defined (_WIN32) && (_MSC_VER >= 1400  && !defined(_AMD64_))
        _localtime32_s(&ltime, (const __time32_t*)&time_stamp);
#else
        cvtime = time_stamp;
        ltime = localtime(&cvtime);
#endif
        if (!list_header) {
            fprintf(stdout, "%s\n", list_title_verbose);
            list_header = 1;
        }
#if _MSC_VER >= 1400  && !defined(_AMD64_)
        fprintf(stdout, "defla %08x %3.3s %02d %02d:%02d %19ld%20ld%6.1f%% %s\n", file_crc, year+ltime.tm_mon*3, ltime.tm_mday, 
                ltime.tm_hour, ltime.tm_min,
                file_info.st_size, file_len, 100*(1.-(float)file_info.st_size/file_len), (char*)orig_fname);
#else
        fprintf(stdout, "defla %08x %3.3s %02d %02d:%02d %19ld%20ld%6.1f%% %s\n", file_crc, year+ltime->tm_mon*3, ltime->tm_mday, ltime->tm_hour, ltime->tm_min,
                file_info.st_size, file_len, 100*(1.-(float)file_info.st_size/file_len), (char*)orig_fname);
#endif
    } else {
        if (!list_header) {
            fprintf(stdout, "%64s\n", list_title);
            list_header = 1;
        }
        fprintf(stdout, "%19ld%20ld%6.1f%% %s\n", file_info.st_size, file_len, 100*(1.-(float)file_info.st_size/file_len), (char*)orig_fname);
    }
}
/**
 * return number of CPU available.
 * if OpenMP is enabled, return omp_get_max_threads
 * 
 * @return int
 */
#ifdef GZIP_PTHREADS
    #include <sched.h> 
    #include <sys/syscall.h> 
    #include <unistd.h>
#endif

int get_cpu_count()
{
    if (opt_dbg_multithreads) {         /* User has specified number of threads to use with "-m" option */
        return num_threads;
    }
#ifndef GZIP_PTHREADS
    #if defined(_OPENMP) && defined(GZIP_PARALLEL)
    return omp_get_max_threads();
    #else
        return 1;
    #endif
#else                               /* Non-OpenMP version is a bit more complex*/
    int cpu_count = 0;
#ifdef __MACH__
    int mib[2];
    size_t len;
#else
    cpu_set_t cpumask; 
    int i;
#endif

#ifdef __MACH__
    mib[0] = CTL_HW;
    mib[1] = HW_NCPU;
    len = sizeof(cpu_count);
    sysctl(mib, 2, &cpu_count, &len, NULL, 0);
    return cpu_count;
#else
    __CPU_ZERO(&cpumask); 
    if(syscall(__NR_sched_getaffinity, 0, sizeof(cpumask), &cpumask) == -1) {
        perror("syscall failed");
        return 1;                       /* assume we are single-CPU */
    }
    for (i=0; i < sizeof(cpumask)*8; i++) {
        if (__CPU_ISSET(i, &cpumask)) {
            cpu_count++;
        }
    }
    return cpu_count;
#endif
#endif
}
/**
 * return 1 if 'filename' already has one of the GZIP suffixes
 * @param filename
 *
 * @return int
 */
int known_suffix(char *filename)
{
        int i;
        char *ptr;

        for (i=0; i<sizeof(gzip_suffixes)/sizeof(char*); i++) {
        size_t str_size;
        ptr = strstr(filename, gzip_suffixes[i]);
        if (ptr != NULL) {
            str_size = strlen(ptr);
            if (str_size == strlen(gzip_suffixes[i])) {
                fprintf(stderr, "%s: %s already has %s suffix -- unchanged\n", prog_name, filename, gzip_suffixes[i]);
                return 1;
            }
        }
        }
    return 0;
}
/**
 * return 1 if "fd" is a regular file
 * 
 * @param fd
 * 
 * @return int
 */
int is_file(int fd)
{
#if _MSC_VER >= 1400  && !defined(_AMD64_)
    struct _stat file_info;
#else
    struct stat file_info;
#endif
#if _MSC_VER >= 1400  && !defined(_AMD64_)
    if (_fstat(fd, &file_info) == 0)
#else
    if (fstat(fd, &file_info) == 0)
#endif
#ifdef _WIN32
        return file_info.st_mode & _S_IFREG;
#else
        return S_ISREG(file_info.st_mode);
#endif
    return 0;
}
/**
 * Print Windows system error code
 * @param text
 */
void error_system(char *text)
{
    print_progname();
#ifdef _WIN32
    fprintf(stderr, " system error code for '%s' %d\n", text, GetLastError());
#else
    fprintf(stderr, "system error code for '%s' %d\n", text, errno);
#endif
}
/**
 * Print out used i/o method to stderr
 */
void print_io_methods()
{
    char *p_m_read, *p_m_write;
    switch (opt_io_read_method) {
        case METHOD_LOW:
            p_m_read = "RL";
            break;
        case METHOD_MMAP:
            p_m_read = "RM";
            break;
        default:
            p_m_read = "RU";
            break;
    }
    switch (opt_io_write_method) {
        case METHOD_LOW:
            p_m_write = "WL";
            break;
        case METHOD_MMAP:
            p_m_write = "WM";
            break;
        default:
            p_m_write = "WU";
            break;
    }
    fprintf(stderr, "%s/%s", p_m_read, p_m_write);
}
